# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized dot_general."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=g-explicit-bool-comparison
# pylint: disable=g-explicit-length-test

import enum
import functools
from typing import Any, Optional, Sequence, Union
from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import calibration
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import int_numerics
from aqt.jax.v2.numerics import no_numerics
import jax
from jax import lax
from jax._src.numpy import lax_numpy
import jax.numpy as jnp
import numpy as onp
from typing_extensions import Self  # for python version < 3.11

dtypes_allowed_for_int32_accum = [jnp.int4, jnp.int8]


class CalibrationMode(enum.Enum):
  """Calibration axis modes."""

  CONTRACTING_AXIS = 1
  REMAINING_AXIS = 2


class DequantMode(enum.Enum):
  """Dequant modes."""

  # Multiply output of dot_general by the transposed scale
  # Compatible with CONTRACTING_AXIS.
  OUTPUT = 1
  # Multiply QTensor.qvalue by untransposed QTensor.scale before
  # dot_general (a.k.a. FakeQuant )
  # Compatible with both CalibrationMode.
  THIS_INPUT = 2
  # Multiply other argument of dot general by appropriately transposed scale.
  # Compatible with REMAINING_AXIS.
  OTHER_INPUT = 3


@utils.flax_slots_dataclass
class Tensor:
  """Configuration of quantization of one tensor or one side of tensor op."""

  quantizer: aqt_quantizer.Quantizer
  # Controls at what value of input tensor should be used.
  # Setting it to True, but not quantizing fwd pass will assert-fail.
  use_fwd_quant: Optional[bool] = utils.static_field()
  # Dequantization mode.
  dequant_mode: DequantMode = utils.static_field()
  # Calibration axis mode.
  calibration_mode: CalibrationMode = utils.static_field()

  @classmethod
  def make(cls, *args, **kwargs) -> Self:
    return tensor_make(*args, **kwargs)


@utils.flax_slots_dataclass
class LocalAqt:
  contraction_axis_shard_count: int = utils.static_field()


def tensor_make(
    bits: Optional[int], preserve_max_val: bool = False
) -> 'Tensor':
  """Makes config.Tensor."""
  if bits is None:
    effective_numerics = no_numerics.NoNumerics()
  else:

    def _dtype_from_bits(bits, pz):
      if 2 <= bits <= 8 and pz:
        if bits == 4:
          return jnp.int4
        else:
          return jnp.int8
      else:
        return None

    pz = False if bits == 1 else True
    dtype = _dtype_from_bits(bits, pz)
    effective_numerics = int_numerics.IntNumerics(
        bits=bits,
        preserve_zero=pz,
        preserve_max_val=preserve_max_val,
        clip=True,
        round=True,
        noise_fn=None,
        clip_gradient=False,  # This can be disabled when using abs-max scaling.
        dtype=dtype,
    )
  quantizer = aqt_quantizer.Quantizer(
      numerics=effective_numerics,
      calib_shared_axes=None,
      scale_stop_grad=True,
      calibration=calibration.AbsMaxCalibration(),
      po2_scale=False,
      context=aqt_quantizer.Context(key=None, train_step=None),
  )
  return Tensor(
      quantizer=quantizer,
      # dtype_x=dtype,
      use_fwd_quant=None,
      dequant_mode=DequantMode.OUTPUT,
      calibration_mode=CalibrationMode.CONTRACTING_AXIS,
  )


def dot_general_raw_make(
    lhs_bits=None,
    rhs_bits=None,
    local_aqt=None,
    jax_scope_name='aqt',
) -> 'DotGeneralRaw':
  """Create quantization configs for input matrices to a matmul."""
  lhs_cfg = tensor_make(lhs_bits)
  rhs_cfg = tensor_make(rhs_bits)

  # Binary uses 0.5 right now.
  if (
      lhs_bits is not None
      and rhs_bits is not None
      and 2 <= lhs_bits <= 8
      and 2 <= rhs_bits <= 8
  ):
    dg_accumulator_dtype = jnp.int32
  else:
    dg_accumulator_dtype = None

  return DotGeneralRaw(
      lhs=lhs_cfg,
      rhs=rhs_cfg,
      dg_accumulator_dtype=dg_accumulator_dtype,
      local_aqt=local_aqt,
      jax_scope_name=jax_scope_name,
  )


def conv_general_dilated_make(
    spatial_dimensions=2,
    lhs_bits: Optional[int] = None,
    rhs_bits: Optional[int] = None,
) -> 'DotGeneralRaw':
  """Create quantization config conv_general_dilated."""
  config = dot_general_raw_make(lhs_bits, rhs_bits)
  # Hardcoding flax assumptions.
  if config.lhs:
    config.lhs.quantizer.calib_shared_axes = list(
        range(1, spatial_dimensions + 2)
    )
  if config.rhs:
    config.rhs.quantizer.calib_shared_axes = list(
        range(0, spatial_dimensions + 2 - 1)
    )
  return config


def dot_general_make(
    lhs_bits: Optional[int] = None,
    rhs_bits: Optional[int] = None,
    bwd_bits: Optional[int] = None,
    use_fwd_quant: bool = True,
    dlhs_local_aqt=None,
    drhs_local_aqt=None,
) -> 'DotGeneral':
  """Create quantization configs for input matrices to a matmul."""
  fwd = dot_general_raw_make(lhs_bits, rhs_bits, jax_scope_name='aqt_fwd')
  dlhs = dot_general_raw_make(
      bwd_bits, bwd_bits, local_aqt=dlhs_local_aqt, jax_scope_name='aqt_dlhs'
  )
  drhs = dot_general_raw_make(
      bwd_bits, bwd_bits, local_aqt=drhs_local_aqt, jax_scope_name='aqt_drhs'
  )
  cfg = DotGeneral(fwd=fwd, dlhs=dlhs, drhs=drhs)

  # Surprising: lhs quantization determines what drhs can do.
  if lhs_bits is not None:
    # Only rhs is accepting MultiTensor.
    cfg.drhs.rhs.use_fwd_quant = use_fwd_quant
  if rhs_bits is not None:
    cfg.dlhs.rhs.use_fwd_quant = use_fwd_quant
  assert cfg.fwd.local_aqt is None, 'local_aqt is not yet supported in fwd.'
  return cfg


# ------------------------------------------------------------------------------


@utils.flax_slots_dataclass
class MultiTensor:
  x: jnp.ndarray
  qx: aqt_tensor.QTensor


@utils.flax_slots_dataclass
class TensorRes:
  """All the things we pass from the forward pass to the backward pass."""

  mt: MultiTensor
  quant_grad: aqt_tensor.GradientFn


@utils.flax_slots_dataclass
class DotGeneralRes:
  lhs: TensorRes
  rhs: TensorRes


def einsum(eqn: str, lhs: jnp.ndarray, rhs: jnp.ndarray, dg=lax.dot_general):
  """A copy of jnp.einsum but without the default jit so as to be injectable."""
  operands, contractions = lax_numpy._default_poly_einsum_handler(  # pylint: disable=protected-access
      eqn, lhs, rhs, einsum_call=True, use_blas=True, optimize='optimal'
  )
  contractions = tuple((a, frozenset(b), c) for a, b, c, *_ in contractions)
  return jax.named_call(lax_numpy._einsum, name=eqn)(  # pylint: disable=protected-access
      operands,
      contractions,
      precision=None,
      preferred_element_type=None,
      _dot_general=dg,
  )


def _scale_trans(x, ca, ba):
  """Transposes x to output dimension order."""
  ca = list(ca)
  ba = list(ba)
  for i in ca:
    assert x.shape[i] == 1
  ra = list(i for i in range(len(x.shape)) if i not in ba + ca)
  x = jnp.transpose(x, ba + ra + ca)
  # TODO(lew): x = jnp.squeeze(x, axis=range(len(ba+ra): len(x.shape))
  shape_ba = x.shape[: len(ba)]
  shape_ra = x.shape[len(ba) : len(x.shape) - len(ca)]
  # Will need to add additional axes (size 1) for the other shape_ra
  x = x.reshape(shape_ba + shape_ra)
  return x


def _lhs_scale_transpose_to_output(
    lhs_scale, dimension_numbers, lhs_shape, rhs_shape
):
  """Transposes lhs_scale to output dimension order."""
  if lhs_scale is None:
    return None
  # The axis order in out is as follows: batch, lhs_ra, rhs_ra
  # - batch axes order is uniquely determined by either lhs_ba or rhs_ba
  # - contraction axes ca disappear from the output
  # - order of the remaining axes (ra) is preserved.
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  qlhs_scale_t = _scale_trans(lhs_scale, lhs_ca, lhs_ba)
  # inserting dummy axes for rhs_ra
  assert len(qlhs_scale_t.shape) == len(lhs_shape) - len(lhs_ca)
  start = len(qlhs_scale_t.shape)
  end = len(rhs_shape) - len(rhs_ca) - len(rhs_ba) + start
  lhs_dummy_axes = range(start, end)
  qlhs_scale_t = jnp.expand_dims(qlhs_scale_t, axis=lhs_dummy_axes)
  return qlhs_scale_t


def _rhs_scale_transpose_to_output(
    rhs_scale, dimension_numbers, lhs_shape, rhs_shape
):
  """Transposes rhs_scale to output dimension order."""
  if rhs_scale is None:
    return None
  del rhs_shape
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  qrhs_scale_t = _scale_trans(rhs_scale, rhs_ca, rhs_ba)
  start = len(rhs_ba)
  end = len(lhs_shape) - len(lhs_ca) - len(lhs_ba) + start
  rhs_dummy_axes = range(start, end)
  qrhs_scale_t = jnp.expand_dims(qrhs_scale_t, axis=rhs_dummy_axes)
  return qrhs_scale_t


def _get_ra(rank: int, ca: Sequence[int], ba: Sequence[int]):
  """Returns the remaining axes."""
  ret = []
  for i in range(rank):
    if i not in list(ca) + list(ba):
      ret.append(i)
  return ret


def _scale_trans_for_other_input(
    x: jax.Array,
    my_ca: Sequence[int],
    my_ba: Sequence[int],
    other_ca: Sequence[int],
    other_ba: Sequence[int],
    other_rank: int,
):
  """Transposes x to other inputs' dimension order."""
  my_ca = list(my_ca)
  my_ba = list(my_ba)
  other_ca = list(other_ca)
  other_ba = list(other_ba)

  # Match the rank.
  if len(x.shape) < other_rank:
    x = x.reshape(list(x.shape) + [1] * (other_rank - len(x.shape)))

  transpose_dim = [-1] * len(x.shape)
  my_axis_mapped = my_ca + my_ba
  other_axis_mapped = other_ca + other_ba
  my_ra = _get_ra(x.ndim, my_ca, my_ba)
  for axis in my_ra:
    assert x.shape[axis] == 1
  for my_axis, other_axis in zip(my_axis_mapped, other_axis_mapped):
    transpose_dim[other_axis] = my_axis

  # Fill unrelated axis with remaining axis.
  ra_idx = 0
  for transpose_dim_idx, transpose_dim_value in enumerate(transpose_dim):
    if transpose_dim_value == -1:
      transpose_dim[transpose_dim_idx] = my_ra[ra_idx]
      ra_idx += 1
  assert ra_idx == len(my_ra)

  # Transpose.
  x = jnp.transpose(x, transpose_dim)

  # Remove redundant axis.
  if len(x.shape) > other_rank:
    for idx in range(len(x.shape), other_rank):
      assert x.shape[idx] == 1
    x = x.reshape(x.shape[:other_rank])

  return x


def _lhs_scale_transpose_for_rhs_input(lhs_scale, dimension_numbers, rhs_shape):
  """Transposes lhs_scale to rhs input dimension order."""
  if lhs_scale is None:
    return None

  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  return _scale_trans_for_other_input(
      lhs_scale, lhs_ca, lhs_ba, rhs_ca, rhs_ba, len(rhs_shape)
  )


def _rhs_scale_transpose_for_lhs_input(rhs_scale, dimension_numbers, lhs_shape):
  """Transposes lhs_scale to rhs input dimension order."""
  if rhs_scale is None:
    return None

  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  return _scale_trans_for_other_input(
      rhs_scale, rhs_ca, rhs_ba, lhs_ca, lhs_ba, len(lhs_shape)
  )


def _get_scale_t(
    qt: aqt_tensor.QTensor,
    transpose_fn: Any,
    dimension_numbers: lax.DotDimensionNumbers,
    lhs_shape: Sequence[int],
    rhs_shape: Sequence[int],
) -> aqt_tensor.QTensor:
  list_scale_t = []
  for scale in qt.scale:
    scale_t = transpose_fn(scale, dimension_numbers, lhs_shape, rhs_shape)
    list_scale_t.append(scale_t)
  return qt.replace(scale_t=list_scale_t)


@utils.flax_slots_dataclass
class DotGeneralRaw:
  """Configuration of quantization of one dot_general without gradient."""

  lhs: Tensor
  rhs: Tensor
  dg_accumulator_dtype: Optional[jnp.dtype] = utils.static_field()
  local_aqt: Optional[LocalAqt] = utils.static_field()
  jax_scope_name: str = utils.static_field()

  # Set it to true in order to train with non-None lhs_qt or rhs_qt.
  # Gradient will still flow into lhs_qt and/or rhs_qt, but it may be incorrect.
  # It is a caller responsibility to NOT update these QTensors
  allow_dummy_gradient_into_qtensor: bool = utils.static_field(default=False)

  @classmethod
  def make(cls, *args, **kwargs) -> Self:
    return dot_general_raw_make(*args, **kwargs)

  @classmethod
  def make_conv_general_dilated(cls, *args, **kwargs) -> Self:
    return conv_general_dilated_make(*args, **kwargs)

  def __call__(
      self,
      lhs: jnp.ndarray,
      rhs: Union[jnp.ndarray, MultiTensor],
      # xhs_qt are used in serving.
      lhs_qt: Optional[aqt_tensor.QTensor],
      rhs_qt: Optional[aqt_tensor.QTensor],
      dimension_numbers: jax.lax.DotDimensionNumbers,
  ):
    """A quantized dot_general function without custom gradient."""
    with jax.named_scope(self.jax_scope_name):
      (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
      # TODO(lew):
      #  - Use qx.value with the int type.
      #  - Handle qx.value with the int type in an optimized way.
      #  - Add a "FQ" case we multiply qx.value*qx.value_scale (not transposed).
      #  - Can we carry untransposed scale and transpose here?
      if isinstance(rhs, MultiTensor):
        # We are in gradient code.
        fwd_quantized = rhs.qx.scale_t is not None and len(rhs.qx.scale_t) == 1  # pytype: disable=attribute-error
        expect_fwd_quantized = self.rhs.use_fwd_quant is not None
        msg = (
            'If fwd is quantized, use_fwd_quant should be either True/False;'
            ' otherwise, use_fwd_quant should be None. Misconfiguration: found'
            f' use_fwd_quant is {self.rhs.use_fwd_quant} in bwd, but fwd'
            f' quantization is {fwd_quantized}.'
        )
        assert fwd_quantized == expect_fwd_quantized, msg
        if self.rhs.use_fwd_quant:
          assert fwd_quantized, msg
          # Cast rhs scales to lhs dtype when multiplying with lhs. This is to
          # avoid an unexpected upcast when rhs is float32 but lhs is float16.
          lhs = lhs * rhs.qx.scale_t[0].astype(lhs.dtype)  # pytype: disable=attribute-error
          # rhs qvalue may be integer. It will be quantized again later, so cast
          # its dtype back to dequant dtype.
          # TODO(yichizh): avoid double quantization and evaluate model quality.
          rhs = rhs.qx.qvalue.astype(rhs.qx.dequant_dtype)  # pytype: disable=attribute-error
        else:
          rhs = rhs.x  # pytype: disable=attribute-error
      else:
        assert self.rhs.use_fwd_quant is None, 'cannot set use_fwd_quant in fwd'

      if self.local_aqt is not None:
        local_aqt = self.local_aqt
        factor = local_aqt.contraction_axis_shard_count  # pytype: disable=attribute-error
        msg = 'Custom calib_shared_axes not implemented for local AQT.'
        assert self.lhs.quantizer.calib_shared_axes is None, msg
        assert self.rhs.quantizer.calib_shared_axes is None, msg

        def factor_reshape(x, ca, ba):
          assert factor is not None
          if len(ca) == 0:
            return x, ca, ba
          shape = list(x.shape)
          ax = ca[0]
          orig_size = shape[ax]
          assert orig_size % factor == 0
          shape[ax] = factor
          shape.insert(ax + 1, orig_size // factor)
          new_ca = [(b + int(b >= ax)) for b in ca]
          assert new_ca[0] == ax + 1
          new_ba = [ax] + [(b + int(b > ax)) for b in ba]
          return x.reshape(shape), new_ca, new_ba

        lhs, lhs_ca, lhs_ba = factor_reshape(lhs, lhs_ca, lhs_ba)
        rhs, rhs_ca, rhs_ba = factor_reshape(rhs, rhs_ca, rhs_ba)

        dimension_numbers = (lhs_ca, rhs_ca), (lhs_ba, rhs_ba)

      assert isinstance(rhs, jnp.ndarray)

      def _compute_qtensor(
          inputs: jnp.ndarray,
          input_qtensor: aqt_tensor.QTensor,
          tensor_cfg: Tensor,
          ndim: int,
          ca: Sequence[int],
          ba: Sequence[int],
          transpose_fn: Any,
      ) -> tuple[aqt_tensor.QTensor, str | aqt_tensor.GradientFn]:
        """Compute qtensor from input or input_qtensor."""
        match tensor_cfg.calibration_mode:
          case CalibrationMode.REMAINING_AXIS:
            calibration_axes = _get_ra(ndim, ca, ba)
          case CalibrationMode.CONTRACTING_AXIS:
            calibration_axes = ca
          case _:
            raise ValueError(
                f'Unknown calibration mode: {tensor_cfg.calibration_mode}'
            )

        if input_qtensor is not None:
          if self.allow_dummy_gradient_into_qtensor:
            # quant_grad might be incorrect here, and should not be used.
            _, quant_grad = tensor_cfg.quantizer.quant(
                inputs, calibration_axes=calibration_axes
            )
          else:
            quant_grad = (
                'Poison. '
                + 'Gradients are not generally expected in serving. '
                + 'Please set allow_dummy_gradient_into_qtensor to True '
                + 'if this is the intended behavior.'
            )
          output_qtensor = input_qtensor
        else:
          output_qtensor, quant_grad = tensor_cfg.quantizer.quant(
              inputs, calibration_axes=calibration_axes
          )
        mode = tensor_cfg.calibration_mode
        if (
            output_qtensor.scale_t is None
            and mode == CalibrationMode.CONTRACTING_AXIS
        ):
          msg = 'scale, scale_t cannot be both unknown'
          assert output_qtensor.scale is not None, msg
          output_qtensor = _get_scale_t(
              qt=output_qtensor,
              transpose_fn=transpose_fn,
              dimension_numbers=dimension_numbers,
              lhs_shape=lhs.shape,
              rhs_shape=rhs.shape,
          )
        return output_qtensor, quant_grad

      lhs_qt, lhs_quant_grad = _compute_qtensor(
          lhs,
          lhs_qt,
          self.lhs,
          lhs.ndim,
          lhs_ca,
          lhs_ba,
          _lhs_scale_transpose_to_output,
      )
      lhs_mt = MultiTensor(x=lhs, qx=lhs_qt)
      lhs_res = TensorRes(mt=lhs_mt, quant_grad=lhs_quant_grad)

      rhs_qt, rhs_quant_grad = _compute_qtensor(
          rhs,
          rhs_qt,
          self.rhs,
          rhs.ndim,
          rhs_ca,
          rhs_ba,
          _rhs_scale_transpose_to_output,
      )
      rhs_mt = MultiTensor(x=rhs, qx=rhs_qt)
      rhs_res = TensorRes(mt=rhs_mt, quant_grad=rhs_quant_grad)

      # TODO(lew): mt.x above should be clipped for clipping calibrations

      out = _qtensor_dot_general(
          lhs_qt, rhs_qt, dimension_numbers, self, jnp.promote_types(lhs, rhs)
      )

      out = out.dequant()

      res = DotGeneralRes(
          lhs=lhs_res,
          rhs=rhs_res,
      )
      if self.local_aqt is not None:
        assert len(lhs_ca) == len(rhs_ca)
        if len(lhs_ca) > 0:
          out = jnp.sum(out, axis=0)
        # We are not supporting local AQT in fwd pass, so no res needed.
        res = None
      return out, res


def _qtensor_dot_general(
    lhs_qt: aqt_tensor.QTensor,
    rhs_qt: aqt_tensor.QTensor,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    cfg: ...,  # DotGeneralRaw,
    # dequant_dtype: DType,
    dequant_dtype: jnp.dtype,
) -> aqt_tensor.QTensor:
  """QTensor lax.dot_general replacement."""

  def _maybe_dequant(
      input_qtensor: aqt_tensor.QTensor, tensor_cfg: Tensor
  ) -> jnp.ndarray:
    if tensor_cfg.dequant_mode == DequantMode.THIS_INPUT:
      output = input_qtensor.dequant()
    else:
      output = input_qtensor.qvalue
    return output

  # Dequantize before the lax dg call if in fake quant mode
  lhs_qin = _maybe_dequant(lhs_qt, cfg.lhs)
  rhs_qin = _maybe_dequant(rhs_qt, cfg.rhs)

  dtype_ms = (
      f'Found {cfg.dg_accumulator_dtype=}, {lhs_qin.dtype=} and'
      f' {rhs_qin.dtype=}. Dot general accumulator dtype can only be'
      ' jnp.int32 when both inputs are int8. Otherwise it is recommended to'
      ' be None to let lax.dot_general automatically decide it.'
  )
  if cfg.dg_accumulator_dtype == jnp.int32:
    assert (
        lhs_qin.dtype in dtypes_allowed_for_int32_accum
        and rhs_qin.dtype in dtypes_allowed_for_int32_accum
    ), dtype_ms

  dtypes_can_be_scaled = [jnp.bfloat16, jnp.float32, jnp.float64]
  if cfg.lhs.dequant_mode == DequantMode.OTHER_INPUT:
    assert rhs_qin.dtype in dtypes_can_be_scaled
    for scale in lhs_qt.scale:
      transposed_scale = _lhs_scale_transpose_for_rhs_input(
          scale, dimension_numbers, rhs_qt.shape
      )
      assert isinstance(transposed_scale, jnp.ndarray)  # make pytype quiet
      rhs_qin = rhs_qin * transposed_scale.astype(rhs_qin.dtype)
  if cfg.rhs.dequant_mode == DequantMode.OTHER_INPUT:
    assert lhs_qin.dtype in dtypes_can_be_scaled
    for scale in rhs_qt.scale:
      transposed_scale = _rhs_scale_transpose_for_lhs_input(
          scale, dimension_numbers, lhs_qt.shape
      )
      assert isinstance(transposed_scale, jnp.ndarray)  # make pytype quiet
      lhs_qin = lhs_qin * transposed_scale.astype(lhs_qin.dtype)

  out = lax.dot_general(
      lhs_qin,
      rhs_qin,
      dimension_numbers=dimension_numbers,
      preferred_element_type=cfg.dg_accumulator_dtype,
      precision=lax.Precision.DEFAULT,
  )
  # TODO(lew): Do we have a correct precision above?
  #   Relevant: https://github.com/google/jax/issues/14022
  out = aqt_tensor.QTensor(
      qvalue=out,
      scale=[],
      scale_t=None,
      dequant_dtype=dequant_dtype,
  )
  assert out.scale is not None  # pytype help

  if cfg.lhs.dequant_mode == DequantMode.OUTPUT:
    out.scale.extend(lhs_qt.scale_t)
  if cfg.rhs.dequant_mode == DequantMode.OUTPUT:
    out.scale.extend(rhs_qt.scale_t)
  return out


@utils.flax_slots_dataclass
class DotGeneral:
  """Configuration of quantization of dot_general and its gradients."""

  fwd: DotGeneralRaw
  dlhs: DotGeneralRaw
  drhs: DotGeneralRaw

  @classmethod
  def make(cls, *args, **kwargs) -> Self:
    return dot_general_make(*args, **kwargs)

  def dg_core(
      self,
      lhs: jnp.ndarray,
      rhs: jnp.ndarray,
      lhs_qt: Optional[aqt_tensor.QTensor],
      rhs_qt: Optional[aqt_tensor.QTensor],
      dimension_numbers: lax.DotDimensionNumbers,
  ):
    """dot_general function with expanded API."""
    msg = 'AQT is not yet optimized to accept quantized types directly. '
    msg += f'lhs.dtype: {lhs.dtype}, rhs.dtype: {rhs.dtype}'
    assert lhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg
    assert rhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg

    self.assert_config_validity()
    out, res = _dg_core(lhs, rhs, lhs_qt, rhs_qt, dimension_numbers, self)
    return out, res

  def __call__(
      self,
      lhs,
      rhs,
      dimension_numbers,
      precision=None,
      preferred_element_type=None,
  ):
    del preferred_element_type
    assert (
        precision is None
    ), f'Precision {precision} requested together with quantization.'

    out, _ = self.dg_core(
        lhs=lhs,
        rhs=rhs,
        lhs_qt=None,
        rhs_qt=None,
        dimension_numbers=dimension_numbers,
    )
    return out

  def assert_config_validity(self: Self):
    """Asserts if configuration is valid."""

    # use_fwd_quant is not enabled when calibration_axis = remaining_axis.
    msg_fwd_quant = (
        'use_fwd_quant should be set to None when remaining axis are used for'
        ' calibration axis.'
    )

    if self.fwd.rhs.calibration_mode == CalibrationMode.REMAINING_AXIS:
      msg_fwd_quant += (
          f'rhs.calibration_mode: {self.fwd.rhs.calibration_mode},'
          f' dlhs use_fwd_quant: {self.dlhs.rhs.use_fwd_quant}'
      )
      assert self.dlhs.rhs.use_fwd_quant is None, msg_fwd_quant

    if self.fwd.lhs.calibration_mode == CalibrationMode.REMAINING_AXIS:
      msg_fwd_quant += (
          f'lhs.calibration_mode: {self.fwd.lhs.calibration_mode},'
          f' drhs use_fwd_quant: {self.drhs.rhs.use_fwd_quant}'
      )
      assert self.drhs.rhs.use_fwd_quant is None, msg_fwd_quant

    # Check valid combination between calibration_mode and dequant_mode
    unsupported_calibration_dequant_pairs = [
        (DequantMode.OUTPUT, CalibrationMode.REMAINING_AXIS),
        (DequantMode.OTHER_INPUT, CalibrationMode.CONTRACTING_AXIS),
    ]
    msg_mode_mismatch = (
        'Unsupported calibration mode - dequant mode combination '
    )
    for (
        dequant_mode,
        calibration_mode,
    ) in unsupported_calibration_dequant_pairs:
      assert not (
          self.fwd.lhs.calibration_mode == calibration_mode
          and self.fwd.lhs.dequant_mode == dequant_mode
      ), (
          msg_mode_mismatch
          + ' for lhs. calibration_mode:'
          f' {self.fwd.lhs.calibration_mode}, dequant_mode:'
          f' {self.fwd.lhs.dequant_mode}'
      )
      assert not (
          self.fwd.rhs.calibration_mode == calibration_mode
          and self.fwd.rhs.dequant_mode == dequant_mode
      ), (
          msg_mode_mismatch
          + ' for rhs. calibration_mode:'
          f' {self.fwd.rhs.calibration_mode}, dequant_mode:'
          f' {self.fwd.rhs.dequant_mode}'
      )


@functools.partial(jax.custom_vjp, nondiff_argnums=(4,))
def _dg_core(
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    lhs_qt: Optional[aqt_tensor.QTensor],
    rhs_qt: Optional[aqt_tensor.QTensor],
    dimension_numbers: lax.DotDimensionNumbers,
    cfg: DotGeneral,
):
  out, _ = _dg_core_vjp_fwd(lhs, rhs, lhs_qt, rhs_qt, dimension_numbers, cfg)
  return out


# When defining a vjp, all traceable variables must be input arguments of
# both the fwd and bwd function.
# The cfg (DotGeneral) contains the key used for stochastic rounding,
# which are traceable dynamic variables. It needs to be an input argument
# to prevent the jax side effect.
def _dg_core_vjp_fwd(
    lhs: jnp.ndarray,
    rhs: jnp.ndarray,
    lhs_qt: Optional[aqt_tensor.QTensor],
    rhs_qt: Optional[aqt_tensor.QTensor],
    dimension_numbers: lax.DotDimensionNumbers,
    cfg: DotGeneral,
):
  """custom_vjp fwd pass."""
  assert (
      lhs.dtype == rhs.dtype
  ), f'Unmatched lhs and rhs dtype: {lhs.dtype} vs {rhs.dtype}'
  ret, res = cfg.fwd(
      lhs,
      rhs,
      lhs_qt,
      rhs_qt,
      dimension_numbers,
  )
  ret = ret.astype(lhs.dtype)
  # We return these values to allow for materialization.
  assert res is not None, 'res cannot be None in fwd pass.'
  qret = (res.lhs.mt.qx, res.rhs.mt.qx)
  return ((ret, qret), (res, cfg))


def _dg_core_vjp_bwd(
    fwd_dimension_numbers: lax.DotDimensionNumbers,
    res: tuple[Optional[DotGeneralRes], DotGeneral],
    g,
):
  """custom_vjp bwd pass."""
  dg_res, cfg = res
  msg = 'dg_res can only be None in 2nd derivative. It is not yet supported.'
  assert dg_res is not None, msg
  g = g[0]  # g[1] is gradient with respect to qret which we are ignoring.

  def ranges_like(*xs):
    start = 0
    for x in xs:
      yield tuple(range(start, start + len(x)))
      start += len(x)

  def grad_dot_general(
      y_res: TensorRes,
      quant_grad: aqt_tensor.GradientFn,
      dg_raw: DotGeneralRaw,
      y_is_lhs,
  ):
    y_ndim = y_res.mt.x.ndim

    (x_ca, y_ca), (x_ba, y_ba) = fwd_dimension_numbers
    if y_is_lhs:
      (y_ca, x_ca) = (x_ca, y_ca)
      (y_ba, x_ba) = (x_ba, y_ba)
    g_ndim = g.ndim - y_ndim + len(x_ba) + 2 * len(x_ca)
    x_ra = tuple(_get_ra(g_ndim, x_ca, x_ba))
    y_ra = tuple(_get_ra(y_ndim, y_ca, y_ba))
    if y_is_lhs:
      g_ba, g_ca, _ = ranges_like(x_ba, y_ra, x_ra)
    else:
      g_ba, _, g_ca = ranges_like(x_ba, x_ra, y_ra)
    dims = ((g_ca, y_ra), (g_ba, y_ba))

    out, _ = dg_raw(g, y_res.mt, None, None, dims)

    x_ca_sorted_by_y = tuple(onp.take(x_ca, onp.argsort(y_ca)))
    out_axes = tuple(onp.argsort(tuple(x_ba) + x_ra + x_ca_sorted_by_y))
    transposed_out = jax.lax.transpose(out, out_axes)
    if quant_grad is not None:
      transposed_out = quant_grad(transposed_out)[0]
    return transposed_out

  dlhs = grad_dot_general(
      dg_res.rhs,
      dg_res.lhs.quant_grad,
      cfg.dlhs,
      False,
  )
  drhs = grad_dot_general(
      dg_res.lhs,
      dg_res.rhs.quant_grad,
      cfg.drhs,
      True,
  )
  # fwd_dimension_numbers are marked as nondiff_argnums instead of returning
  # None as grad to it. This is because it is a tuple of Python integers
  # that cannot be traced by Jax.
  return (dlhs, drhs, None, None, None)


_dg_core.defvjp(_dg_core_vjp_fwd, _dg_core_vjp_bwd)


def make_dot_general(dg: Optional[DotGeneral]):
  # TODO(lew): call warnings.warn("Deprecated")
  if dg is None:
    return jax.lax.dot_general
  else:
    return dg


def assert_config_validity(cfg):
  # TODO(lew): call warnings.warn("Deprecated")
  cfg.assert_config_validity()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Creates a sweep over all the leaderboard models.

Model size: full model
Cadence: 100K, 150K
Sweep:
Parts: {encoder, decoder, encoder + decoder}
Inputs: {weights}
Sparsity type: {1:4, 2:4}
"""

import copy

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.sparsity import full_model_bfloat16_sparse_dense_weights_only_structured_nm


import ml_collections


def get_config():
  """Returns sweep configuration (see module docstring)."""
  sweep_config = ml_collections.ConfigDict()
  base_configs = [
      full_model_bfloat16_sparse_dense_weights_only_structured_nm,
  ]
  configs = []
  for base_config_file in base_configs:
    for n, m in zip([32], [32]):
      base_config = base_config_file.get_config(n, m)
      base_config.mlp_block.dense_1.weight_sparsity.structure_decay = True
      base_config.mlp_block.dense_2.weight_sparsity.structure_decay = True
      config = copy.deepcopy(base_config)
      config.metadata.hyper_str = f"{config.metadata.hyper_str}_N({n})_M({m})"
      configs.append(config)
  sweep_config.configs = configs
  return sweep_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to generate ConfigDict instances.

This is the shared part of config schema without model-specific components.
Most users will just create a ConfigDict instance with 'get_config' and then
override its parameters to specialize the configuration.
"""

from typing import List, Optional, Union

import ml_collections
from ml_collections import config_dict


def make_reference(
    config: ml_collections.ConfigDict,
    field: str,
) -> Union[ml_collections.ConfigDict, ml_collections.FieldReference]:
  """Returns a reference to a field for wiring up a config dict.

  The returned reference is "one-way": a change to the original field value
  propagates to the reference, but a change to the reference does not propagate
  back up.

  This works recursively. If config['field'] is itself a ConfigDict (or a tree
  of ConfigDicts) then an identical tree is constructed with all the leaves
  referring to original leaves. Internal nodes are *not* references.

  See config_schema_test.MakeReferenceRecursiveTest for example usage.

  Args:
    config: A ConfigDict instance.
    field: The name of a field contained in 'config'.

  Returns:
    If config['field'] is a scalar, returns a one-way FieldReference to
      config['field']. If config['field'] is a ConfigDict, returns a new
      ConfigDict whose fields are references to the fields in config['field'].
  """
  field_value = config[field]
  if isinstance(field_value, ml_collections.ConfigDict):
    new_dict = ml_collections.ConfigDict()
    for subfield in field_value:
      new_dict[subfield] = make_reference(field_value, subfield)
    return new_dict
  else:
    return ml_collections.FieldReference(config.get_ref(field))


def set_default_reference(child_config: ml_collections.ConfigDict,
                          parent_config: ml_collections.ConfigDict,
                          field: Union[str, List[str]],
                          *,
                          parent_field: Optional[str] = None) -> None:
  """Sets fields in a child config to be references to fields in a parent.

  When all parameters are provided and field is a single string, this
  is equivalent to
  child_config[field] =
    make_reference(parent_config[parent_field])

  See documentation on make_reference for the full implications.

  If parent_field is not specified, it is taken to be the same as 'field'.
  This is useful for the common case where the name of a field is the same
  in the parent and child.

  If 'field' is a list of strings, that is equivalent to calling this
  function separately with each specific string. Useful for creating many
  references at once.

  See config_schema_test.SetDefaultReferenceTest for example usage.

  Args:
    child_config: The child ConfigDict.
    parent_config: The parent ConfigDict.
    field: Either the name of a field or a list of field names.
    parent_field: The name of the parent field in 'parent_config'.
      If None (the default), the parent field name is assumed to be the same
      as the child field name.
  """
  if isinstance(field, list):
    for subfield in field:
      set_default_reference(
          child_config=child_config,
          field=subfield,
          parent_config=parent_config,
          parent_field=parent_field)
  else:
    if parent_field is None:
      parent_field = field
    child_config[field] = make_reference(parent_config, parent_field)


# Functions returning placeholders are marked with _ph suffix are a device
# to increase code reability in this file. Their intent is to reduce large
# amount of repetition and getting the type closer to the colon.
float_ph = lambda: config_dict.placeholder(float)
int_ph = lambda: config_dict.placeholder(int)
str_ph = lambda: config_dict.placeholder(str)
bool_ph = lambda: config_dict.placeholder(bool)


def get_dense_config(
    parent_config: ml_collections.ConfigDict) -> ml_collections.ConfigDict:
  """Creates a ConfigDict corresponding to aqt.flax_layers.DenseAqt.HParams."""
  config = ml_collections.ConfigDict()
  set_default_reference(
      config,
      parent_config,
      [
          "weight_prec",
          "weight_quant_granularity",
          "quant_type",
          "quant_act",
          "weight_half_shift",
          "act_sparsity",
          "weight_sparsity",
      ])
  config.lock()
  return config


def get_conv_config(
    parent_config: ml_collections.ConfigDict) -> ml_collections.ConfigDict:
  """Creates a ConfigDict corresponding to aqt.flax_layers.ConvAqt.HParams."""
  config = ml_collections.ConfigDict()
  set_default_reference(config, parent_config, [
      "weight_prec", "weight_quant_granularity", "quant_type", "quant_act",
      "weight_half_shift",
  ])
  config.lock()
  return config


def get_fp_quant_config() -> ml_collections.ConfigDict:
  config = ml_collections.ConfigDict({
      "fp_spec": get_fp_config(),
      "is_scaled": bool_ph(),
  })
  config.lock()
  return config


def get_fp_config() -> ml_collections.ConfigDict:
  config = ml_collections.ConfigDict({
      "exp_min": int_ph(),
      "exp_max": int_ph(),
      "sig_bits": int_ph()
  })
  config.lock()
  return config


def get_sparse_config(use_unstructured: bool) -> ml_collections.ConfigDict:
  """Returns a sparse ConfigDict based on sparsity type argument."""
  prune_rate = float_ph() if use_unstructured else (int_ph(), int_ph())
  config = ml_collections.ConfigDict({
      "type": str_ph(),
      "prune_rate": prune_rate,
      "smallest": bool_ph(),
      "order": str_ph(),
      "absolute": bool_ph(),
      "structure_decay": bool_ph(),
      "sparse_ste": bool_ph()
  })
  config.lock()
  return config


# TODO(shivaniagrawal): base config should be more generic and only model
# specific configs should be updated.
def get_base_config(
    use_auto_acts: bool,
    fp_quant: bool,
    use_unstructured: bool = False,
) -> ml_collections.ConfigDict:
  """Return a base ConfigDict for AQT; does not have model specific fields."""
  if use_auto_acts:
    bounds = ml_collections.ConfigDict({
        "initial_bound": float_ph(),
        "stddev_coeff": float_ph(),
        "absdev_coeff": float_ph(),
        "mix_coeff": float_ph(),
        "reset_stats": bool_ph(),
        "ema_coeff": float_ph(),
        "use_cams": bool_ph(),
        "exclude_zeros": bool_ph(),
        "use_mean_of_max": bool_ph(),
        "granularity": str_ph()
    })
  else:
    bounds = float_ph()
  if fp_quant:
    prec = get_fp_quant_config()
  else:
    prec = int_ph()
  sparsity = get_sparse_config(use_unstructured)
  base_config = ml_collections.ConfigDict({
      "metadata": {
          "description": "Base configuration",
          "hyper_str": str_ph(),
      },
      "weight_decay": float_ph(),
      "activation_bound_update_freq": int_ph(),
      "activation_bound_start_step": int_ph(),
      "prec": prec,
      "sparsity": sparsity,
      "half_shift": bool_ph(),
      "quant_type": str_ph(),
      "quant_act": {
          "bounds": bounds,
          # TODO(shivaniagrawal): The input distribution is really an intrinsic
          # model property and shouldn't be part of the model configuration.
          # Update the hparam dataclasses to eliminate the input_distribution
          # field and then delete this.
          "input_distribution": "symmetric"
      },
      "weight_quant_granularity": str_ph()
  })

  set_default_reference(
      base_config, base_config, "weight_prec", parent_field="prec")
  set_default_reference(base_config.quant_act, base_config, "prec")
  set_default_reference(
      base_config, base_config, "act_sparsity", parent_field="sparsity")
  set_default_reference(
      base_config, base_config, "weight_sparsity", parent_field="sparsity")

  set_default_reference(
      base_config, base_config, "weight_half_shift", parent_field="half_shift")
  set_default_reference(base_config.quant_act, base_config, "half_shift")

  return base_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for retrieving state values from flax models."""

from typing import Any, Iterable, Mapping

from flax.metrics import tensorboard
import numpy as np


def get_state_dict_summary(state_dict: Mapping[str, Any],
                           keys: Iterable[str]) -> Mapping[str, Any]:
  """Gets tagged distributions from state dictionary.

  Retrieves values for specified keys from state_dict if recorded.

  Args:
     state_dict: model state as dictionary.
     keys: The keys to look up in state_dict.

  Returns:
    dictionary containing state values corresponding to keys if recorded.

  """

  return _get_state_dict_summary_recursive(state_dict, keys, '', {})


def _get_state_dict_summary_recursive(parent_module, keys,
                                      path_to_parent_module, summary):
  for module_name, module in parent_module.items():
    path_to_module = f'{path_to_parent_module}/{module_name}'
    for key in keys:
      if key == module_name:
        summary[path_to_module] = module
    if hasattr(module, 'items'):  # Test if module is dictionary-like
      _get_state_dict_summary_recursive(module, keys, path_to_module, summary)
  return summary


def write_state_dict_summaries_to_tb(
    state_dict_summary_all: Mapping[str, Any],
    train_summary_writer: tensorboard.SummaryWriter,
    state_dict_summary_freq: int, step: int):
  """Write out state dict summaries to Tensorboard."""
  for key, val in state_dict_summary_all.items():
    val = np.asarray(val).astype(np.float32)
    steps_recorded = val.shape[0]
    if key.endswith('bounds'):
      # For bounds, also record mean bound as a scalar.
      for i in range(0, steps_recorded, state_dict_summary_freq):
        train_summary_writer.scalar(f'{key}_mean', val[i:i + 1].mean(), step)

    for i in range(0, steps_recorded, state_dict_summary_freq):
      train_summary_writer.histogram(
          key, val[i:i + 1], step=step - steps_recorded + i, bins=100)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test utility for aqt.jax."""

import math
import typing
from typing import Any, Mapping, Union
from aqt.jax_legacy.jax import quantization
from aqt.jax_legacy.jax import stats
from jax import config
import jax.numpy as jnp
import numpy as np


def assert_all_close_prec(
    exact: jnp.ndarray,
    res: jnp.ndarray,
    prec: Union[None, int, quantization.QuantOps.FloatQuant],
):
  """Assert that |exact - res| < 2^-prec.

  Args:
    exact: The exact result of the computation.
    res: The computed result.
    prec: The quantization precision.
  """
  if prec is None:
    np.testing.assert_allclose(exact, res)
  elif isinstance(prec, quantization.QuantOps.FloatQuant):
    prec = typing.cast(quantization.QuantOps.FloatQuant, prec)
    # relative-error = |x' - x| / |x|
    #  Assume x' & x differ only in significant bits
    #
    #  => exp(x') = exp(x)
    #     2^exp(x) | mant(x') - mant(x) | / 2^exp(x) mant(x)
    #  1 <= mant(x) < 2
    #       mant(x') - mant(x) < 1.xx...x111111... - 1.xx...x < 2^-sig_bits
    #
    #  where xx...x has length equal to the number of significant bits s.t.
    #  |x' - x| / |x| = 2^-sig_bits / |x| < 2^-sig_bits
    np.testing.assert_allclose(exact, res, rtol=2**-prec.fp_spec.sig_bits)
  else:
    # decimal precision corresponding to prec [number of bits]
    dec_prec = prec * math.log(2, 10)
    atol = math.pow(2, -dec_prec)
    # |a - b| <= abs(atol + rtol * abs(b))
    np.testing.assert_allclose(exact, res, rtol=0, atol=atol)


def _iterate_stats_recursive(state1, state2, layer_name):
  """Yields a tensor for each equivalent leaf in two states."""
  if isinstance(state1, stats.Stats):
    stats1 = typing.cast(stats.Stats, state1)
    stats2 = typing.cast(stats.Stats, state2)
    yield layer_name, stats1.mean, stats2.mean
    yield layer_name, stats1.mean_abs, stats2.mean_abs
    yield layer_name, stats1.mean_sq, stats2.mean_sq
  elif hasattr(state1, 'items'):
    for module_name, state1_module in state1.items():
      state2_module = state2[module_name]
      yield from _iterate_stats_recursive(state1_module, state2_module,
                                          f'{layer_name}/{module_name}')
  else:
    yield layer_name, state1, state2


def _iterate_stats(state1, state2):
  yield from _iterate_stats_recursive(state1['get_bounds'],
                                      state2['get_bounds'], '')
  if 'stats_tag' in state1:
    yield from _iterate_stats_recursive(state1['stats_tag'],
                                        state2['stats_tag'], '')


def assert_stats_are_equal(state1: Any, state2: Any):
  """Asserts that the activation statistics in two Flax states are almost equal."""

  for layer_name, state1_stats, state2_stats in _iterate_stats(state1, state2):
    # The tolerance was chosen empirically to make the tests pass reliably for a
    # 3-layer model. Possibly the tolerance has to be high because of floating
    # point accumulated error over 3 layers and because masked-out tokens still
    # exerts a tiny but non-zero padding on statistics since they are masked out
    # in the attention layers by subtracting a large but not infinite negative
    # value from their position in the  Q*K output before taking a softmax to
    # get attention weights.
    np.testing.assert_allclose(
        state1_stats,
        state2_stats,
        err_msg=f'Stats changed for layer {layer_name}',
        atol=.01)


def assert_stats_are_unequal(state1: Mapping[str, Any], state2: Mapping[str,
                                                                        Any]):
  """Asserts that in at least one layer, the statistics in two Flax states are unequal."""

  for _, state1_stats, state2_stats in _iterate_stats(state1, state2):
    if not np.allclose(state1_stats, state2_stats, atol=.01):
      return  # We have found a layer where the stats are different

  # If we reached here, all layers have the same stats
  assert False, 'Activation statistics are the same in all layers'


def configure_jax():
  config.update('jax_numpy_rank_promotion', 'raise')



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for matmul."""

import copy

from absl.testing import parameterized
from aqt.common import aqt_config
from aqt.test import aqt_test_shared_base
import numpy as np
import tensorflow.compat.v1 as tf


def test_stats_config():
  """Config base for all the experiments."""
  return aqt_config.StatsConfig(
      ema_update_count=1,  # single update overwrites the stats
      update_count_prior=0,  # easier equations
      share_stats_axes=[0, 1],  # one stat per whole tensor
      tpu_cross_replica_sum=False,  # no cross-tpu reduction
      filter_zeros=False,  # on default zeros are a number like any other
  )


def calibration_config(const_coeff: float) -> aqt_config.CalibrationConfig:
  return aqt_config.CalibrationConfig(const_bound_coeff=const_coeff)


def _schedule_config(
    bits,
    const_bound_coeff,
    share_stats_axes,
    freeze_scale_at_begin=True,
) -> aqt_config.AqtScheduleConfig:
  """Creates schedule config with dynamic quantization."""
  iqc = aqt_config.IntQuantConfig(bits=bits)
  cc = aqt_config.CalibrationConfig(const_bound_coeff=const_bound_coeff)
  tc = aqt_config.AqtTensorConfig(
      quant_config=iqc,
      calibration_config=cc,
      freeze_scale_at_begin=freeze_scale_at_begin)
  sc = aqt_config.StatsConfig(
      ema_update_count=1,
      share_stats_axes=list(share_stats_axes),
      update_count_prior=0,
      tpu_cross_replica_sum=False)
  return aqt_config.AqtScheduleConfig(sc, [tc])


def config_from_schedule(schedule):
  """Generates a schedule config from [(start, end, bits), ...]."""
  tensor_configs = []
  for start, end, bits in schedule:
    int_quant_config = aqt_config.IntQuantConfig(bits, preserve_zero=False)

    tensor_config = aqt_config.AqtTensorConfig(
        freeze_scale_at_begin=True,
        quant_config=int_quant_config,
        begin_at_event=start,
        end_at_event=end,
        calibration_config=calibration_config(1.0))

    tensor_configs.append(tensor_config)
  return aqt_config.AqtScheduleConfig(test_stats_config(), tensor_configs)


class MatmulTest(tf.test.TestCase, parameterized.TestCase):
  """Base class for testing aqt_matmul in TF and Jax."""

  def constant(self, x):
    raise NotImplementedError

  def matmul(self, config, lhs_shape, rhs_shape, name="aqt"):
    raise NotImplementedError

  def matmul_apply(self, module, lhs, rhs, train=True, keep_stats=False):
    raise NotImplementedError

  def matmul_unquantized(self, lhs, rhs):
    raise NotImplementedError

  def gradients(self, fwd_func, x, w, reduce_sum=False):
    raise NotImplementedError

  def exact_int8_matmul_example(
      self,
      lhs_use_quantized_variable=False,
      rhs_use_quantized_variable=False,
      name="aqt",
      use_float_config=False,
      use_grad_quantization=False,
  ):
    """Creates lhs/rhs tensors, their corresponding AQT configs, and matmul op."""
    lhs_config, lhs, rhs_config, rhs = aqt_test_shared_base.exact_int8_example(
        lhs_shape=(3, 2),
        rhs_shape=(2, 2),
        lhs_share_stats_axes=[0, 1],
        rhs_share_stats_axes=[0, 1],
        lhs_use_quantized_variable=lhs_use_quantized_variable,
        rhs_use_quantized_variable=rhs_use_quantized_variable)

    if use_float_config:
      lhs_config.tensor_configs[0].quant_config = aqt_config.FloatConfig()
    if use_grad_quantization:
      grad_config = copy.deepcopy(lhs_config)
      grad_config.tensor_configs[0].freeze_scale_at_begin = False
      grad_config.stats_config.ema_update_count = 1
    else:
      grad_config = None

    lhs = self.constant(lhs)
    rhs = self.constant(rhs)

    config = aqt_config.AqtMatmulConfig(lhs_config, rhs_config, grad_config)
    mm = self.matmul(config, lhs.shape, rhs.shape, name)

    return mm, lhs, rhs

  def test_matmul_none(self):
    """Ensures no quantization and float config give aqt_matmul = {tf, jnp}.matmul."""
    no_quant_config = aqt_config.AqtScheduleConfig(test_stats_config(), [])
    float_config_tc = aqt_config.AqtTensorConfig(
        freeze_scale_at_begin=True,
        quant_config=aqt_config.FloatConfig(),
        calibration_config=calibration_config(1))
    float_config = aqt_config.AqtScheduleConfig(test_stats_config(),
                                                [float_config_tc])

    lhs = np.random.uniform(-1.0, 1.0, size=(2, 3)).astype(np.float32)
    lhs = self.constant(lhs)

    rhs = np.random.uniform(-1.0, 1.0, size=(3, 4)).astype(np.float32)
    rhs = self.constant(rhs)

    config = aqt_config.AqtMatmulConfig(no_quant_config, no_quant_config)
    mm_no_quant = self.matmul(config, lhs.shape, rhs.shape, "no_quant")
    config = aqt_config.AqtMatmulConfig(float_config, float_config)
    mm_float = self.matmul(config, lhs.shape, rhs.shape, "float")

    no_quant_ret = self.matmul_apply(mm_no_quant, lhs, rhs)
    float_config_ret = self.matmul_apply(mm_float, lhs, rhs)
    expected_ret = self.matmul_unquantized(lhs, rhs)

    self.assertAllEqual(no_quant_ret, expected_ret)
    self.assertAllEqual(float_config_ret, expected_ret)

  def basic_quant_example(self):
    """Returns hand-computed examples."""
    lhs_config = _schedule_config(3, 7, [0, 1])
    lhs = self.constant(
        np.array(
            [
                [-8, 4.01, 4.01],  #
                [-5.99, 0.01, -4.01],
            ],))
    qlhs = self.constant(
        np.array(
            [
                [-6, 4, 4],  #
                [-6, 0, -4]
            ],))

    # Representable values: -1, 0, 1
    rhs_config = _schedule_config(2, 1.5, [0, 1])
    rhs = self.constant(
        np.array(
            [
                [-3, 0.99],  #
                [-0.99, 0],
                [-0.01, 2]
            ],))
    qrhs = self.constant(
        np.array(
            [
                [-1, 1],  #
                [-1, 0],
                [0, 1]
            ],))

    config = aqt_config.AqtMatmulConfig(lhs_config, rhs_config)

    return config, lhs, qlhs, rhs, qrhs

  def test_basic_matmul(self):
    """Test on a hand-computed example."""
    config, lhs, qlhs, rhs, qrhs = self.basic_quant_example()

    mm = self.matmul(config, lhs.shape, rhs.shape)

    actual = self.matmul_apply(mm, lhs, rhs)
    expected = self.matmul_unquantized(qlhs, qrhs)

    self.assertAllEqual(expected, actual)

  @parameterized.parameters([dict(lhs_float=True), dict(lhs_float=False)])
  def test_float_config_basic_matmul(self, lhs_float):
    """Float config test on hand-computed exmaple."""
    config, lhs, qlhs, rhs, qrhs = self.basic_quant_example()

    if lhs_float:
      config.lhs.tensor_configs[0].quant_config = aqt_config.FloatConfig()
    else:
      config.rhs.tensor_configs[0].quant_config = aqt_config.FloatConfig()

    mm = self.matmul(config, lhs.shape, rhs.shape)

    actual = self.matmul_apply(mm, lhs, rhs)
    if lhs_float:
      qlhs = lhs  # lhs is not quantized
    else:
      qrhs = rhs  # rhs is not quantized
    expected = self.matmul_unquantized(qlhs, qrhs)

    self.assertAllEqual(expected, actual)

  @parameterized.named_parameters(
      aqt_test_shared_base.generate_unaligned_schedule_intervals())
  def test_unaligned_schedule_intervals(self, lhs_intervals, rhs_intervals):
    """Raises error if lhs and rhs config schedules are not aligned."""
    bits = 8
    lhs_intervals = [(start, stop, bits) for start, stop in lhs_intervals]
    rhs_intervals = [(start, stop, bits) for start, stop in rhs_intervals]
    lhs_config = config_from_schedule(lhs_intervals)
    rhs_config = config_from_schedule(rhs_intervals)
    config = aqt_config.AqtMatmulConfig(lhs_config, rhs_config)

    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "intervals do not match|config len"):
      self.matmul(config, (1, 1), (1, 1))

  def test_vars_dont_kill_grads(self):
    """Ensures quantized variables do not change gradients."""
    mm, lhs, rhs = self.exact_int8_matmul_example()
    unsaved_matmul = lambda lhs, rhs: self.matmul_apply(mm, lhs, rhs)

    mm_q, _, _ = self.exact_int8_matmul_example(True, True, "with_var")
    saved_matmul = lambda lhs, rhs: self.matmul_apply(mm_q, lhs, rhs)

    saved_grads = self.gradients(saved_matmul, lhs, rhs)
    unsaved_grads = self.gradients(unsaved_matmul, lhs, rhs)

    zipped_grads = zip(saved_grads, unsaved_grads)
    for actual_grad, expected_grad in zipped_grads:
      actual = actual_grad
      expected = expected_grad

      self.assertAllEqual(actual, expected)

  @parameterized.parameters([
      dict(lhs_use_quantized_variable=True),
      dict(lhs_use_quantized_variable=False)
  ])
  def test_vars_over_inputs_at_inference(self, lhs_use_quantized_variable):
    """Ensures quantized variables are used if TQ.use_quantized_variable=True at inference."""
    rhs_use_quantized_variable = not lhs_use_quantized_variable
    mm, tf_lhs, tf_rhs = self.exact_int8_matmul_example(
        lhs_use_quantized_variable, rhs_use_quantized_variable)
    mm_train, _, _ = self.exact_int8_matmul_example(
        lhs_use_quantized_variable=False,
        rhs_use_quantized_variable=False,
        name="no_quantize")

    actual = self.matmul_apply(
        mm, tf_lhs, tf_rhs, train=False, keep_stats=True)
    expected = np.zeros_like(actual)
    # Rely on zero initialization for variables as opposed to non-zero inputs.
    self.assertAllEqual(actual, expected)

    # But if train then use input instead.
    actual = self.matmul_apply(
        mm, tf_lhs, tf_rhs, train=True, keep_stats=True)

    expected = self.matmul_apply(
        mm_train, tf_lhs, tf_rhs, keep_stats=True)
    self.assertAllEqual(actual, expected)

  def test_float_config_not_save_quantized_var(self):
    mm, lhs, rhs = self.exact_int8_matmul_example(
        lhs_use_quantized_variable=True, rhs_use_quantized_variable=True,
        use_float_config=True)

    actual = self.matmul_apply(mm, lhs, rhs, train=False)
    expected = np.zeros_like(actual)
    # Although lhs config sets use_quantized_variable to True, lhs has
    # a float config, and thus it uses zero-initialized quantized var.
    self.assertAllEqual(actual, expected)

  def test_exact_grads(self):
    """Ensures both quantized and unquantized conv op emit the same gradients."""
    mm, lhs, rhs = self.exact_int8_matmul_example()

    aqt_mm = lambda lhs, rhs: self.matmul_apply(mm, lhs, rhs)
    aqt_mm_grads = self.gradients(aqt_mm, lhs, rhs)

    mm_exact = self.matmul_unquantized
    mm_grads = self.gradients(mm_exact, lhs, rhs)

    for aqt_grad, tf_grad in zip(aqt_mm_grads, mm_grads):
      actual = aqt_grad
      expected = tf_grad

      self.assertAllEqual(actual, expected)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for config_schema.py."""

import json

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.utils import config_schema_utils
import ml_collections


class MakeReferenceRecursiveTest(absltest.TestCase):

  def test_scalar_fields(self):
    config = ml_collections.ConfigDict({'parent_field': 1})
    config.child_field = config_schema_utils.make_reference(
        config, 'parent_field')

    # 'child_field' is a reference to 'parent_field'. Changes to
    # 'parent_field' propagate to 'child_field'.
    self.assertEqual(config.parent_field, 1)
    self.assertEqual(config.child_field, 1)

    config.parent_field = 2
    self.assertEqual(config.parent_field, 2)
    self.assertEqual(config.child_field, 2)

    # But changes to 'child_field' to NOT propagate back up to
    # 'parent_field'.
    config.child_field = 3
    self.assertEqual(config.parent_field, 2)
    self.assertEqual(config.child_field, 3)

    config.parent_field = 4
    self.assertEqual(config.parent_field, 4)
    # Reference is broken after 'child_field' was overridden earlier.
    self.assertEqual(config.child_field, 3)

  def test_nested_fields(self):
    config = ml_collections.ConfigDict({'parent': {'x': 1}})
    config.child = config_schema_utils.make_reference(config, 'parent')

    # In this case, 'config.child.x' is a reference to 'config.parent.x', but
    # note that 'config.child' is NOT a reference to 'config.parent'!

    self.assertEqual(config.parent.x, 1)
    self.assertEqual(config.child.x, 1)

    config.parent.x = 2
    self.assertEqual(config.parent.x, 2)
    self.assertEqual(config.child.x, 2)

    config.parent = ml_collections.ConfigDict({'x': 3})
    self.assertEqual(config.parent.x, 3)
    # In this case, config.parent is a new Python object unrelated to the old
    # config.parent. Since config.child is a reference to the old config.parent,
    # it has no connection to the new config.parent.
    self.assertEqual(config.child.x, 2)

    # However, this works as intended since the 'update' function assigns new
    # values to existing leaf nodes, preserving the reference structure between
    # parent and child internal nodes. Using this syntax is recommended for
    # updating many fields at once.
    config = ml_collections.ConfigDict({'parent': {'x': 1, 'y': 'hello'}})
    config.child = config_schema_utils.make_reference(config, 'parent')
    config.parent.update({'x': 3, 'y': 'goodbye'})
    self.assertEqual(config.parent.x, 3)
    self.assertEqual(config.parent.y, 'goodbye')

    self.assertEqual(config.child.x, 3)
    self.assertEqual(config.child.y, 'goodbye')


class SetDefaultReferenceTest(absltest.TestCase):

  def test_when_child_field_is_list(self):
    # Test when 'field' parameter of set_default_reference is a list
    # of specific fields. We expect a new reference to be created for each
    # element in the list.
    parent = ml_collections.ConfigDict({'x': 1, 'y': 2, 'z': 3})
    child = ml_collections.ConfigDict()
    config_schema_utils.set_default_reference(child, parent, ['x', 'y'])
    self.assertEqual((parent.x, parent.y), (1, 2))
    self.assertEqual((child.x, child.y), (1, 2))

    parent.y = 5
    self.assertEqual((parent.x, parent.y), (1, 5))
    self.assertEqual((child.x, child.y), (1, 5))

    child.y = 10
    self.assertEqual((parent.x, parent.y), (1, 5))
    self.assertEqual((child.x, child.y), (1, 10))

  def test_reference_to_self(self):
    # Test adding a new field to a configdict which is a reference to an
    # existing field in the same configdict instance.
    config = ml_collections.ConfigDict({'parent': 1})
    config_schema_utils.set_default_reference(
        config, config, 'child', parent_field='parent')
    self.assertEqual(config.child, 1)
    self.assertEqual(config.parent, 1)

    config.parent = 5
    self.assertEqual(config.parent, 5)
    self.assertEqual(config.child, 5)

    config.child = 10
    self.assertEqual(config.parent, 5)
    self.assertEqual(config.child, 10)


class BaseConfigTest(parameterized.TestCase):

  @parameterized.parameters(dict(use_auto_acts=True), dict(use_auto_acts=False))
  def test_precision_propagates(self, use_auto_acts):
    config = config_schema_utils.get_base_config(use_auto_acts, fp_quant=False)

    # Set the global precision to 4 bits.
    config.prec = 4
    # Set the global half_shift flag to False
    config.half_shift = False
    # Test that this sets the weight and activation to 4 as well.
    self.assertEqual(config.weight_prec, 4)
    self.assertEqual(config.quant_act.prec, 4)
    # Test that this sets the weight_half_shift and act half_shift to False
    self.assertEqual(config.weight_half_shift, False)
    self.assertEqual(config.quant_act.half_shift, False)

    # Set the global precision to None, checks whether referencing to None
    # works well.
    config.prec = None
    # Test that this sets the weight and activation to None as well.
    self.assertIsNone(config.weight_prec, None)
    self.assertIsNone(config.quant_act.prec, None)

  @parameterized.parameters(dict(use_auto_acts=True), dict(use_auto_acts=False))
  def test_fp_precision_propagates(self, use_auto_acts):
    config = config_schema_utils.get_base_config(use_auto_acts, fp_quant=True)

    config.prec.is_scaled = False
    # Set the global precision to 4 bits.
    config.prec.fp_spec.update({'exp_min': -3, 'exp_max': 5, 'sig_bits': 2})

    expected_prec_dict = {
        'is_scaled': False,
        'fp_spec': {
            'exp_min': -3,
            'exp_max': 5,
            'sig_bits': 2
        }
    }
    # Test that this sets the weight and activation to 4 as well.
    self.assertEqual(config.weight_prec.to_dict(), expected_prec_dict)
    self.assertEqual(config.quant_act.prec.to_dict(), expected_prec_dict)

  @parameterized.parameters(
      dict(use_unstructured=True), dict(use_unstructured=False))
  def test_sparsity_propagates(self, use_unstructured):
    config = config_schema_utils.get_base_config(
        use_auto_acts=True, fp_quant=False, use_unstructured=use_unstructured)

    prune_rate = 0.2 if use_unstructured else (1, 2)

    expected_sparse_dict = {
        'type': 'unstructured',
        'prune_rate': prune_rate,
        'smallest': True,
        'absolute': False,
        'order': 'C',
        'structure_decay': False,
        'sparse_ste': False
    }

    config.sparsity.update(expected_sparse_dict)

    # Test that this sets the weight and activation to 4 as well.
    self.assertEqual(config.weight_sparsity.to_dict(), expected_sparse_dict)
    self.assertEqual(config.act_sparsity.to_dict(), expected_sparse_dict)

  def test_auto_acts_parameter(self):
    # If use_auto_acts is False, then the bounds should be a single scalar that
    # specifies the fixed bound; 'None' by default.
    config = config_schema_utils.get_base_config(
        use_auto_acts=False, fp_quant=False)
    self.assertIsNone(config.quant_act.bounds)
    # If use_auto_acts is True, it should have the same structure as the
    # GetBounds.Hyper dataclass.
    config = config_schema_utils.get_base_config(
        use_auto_acts=True, fp_quant=False)
    self.assertIn('initial_bound', config.quant_act.bounds)

    # Because the config dict is locked, it shouldn't be possible to change it
    # back to fixed bounds if it was created with use_auto_acts=True.
    with self.assertRaises(TypeError):
      config.quant_act.bounds = 1.0

  @parameterized.parameters(
      dict(use_auto_acts=True, fp_quant=False),
      dict(use_auto_acts=False, fp_quant=False),
      dict(use_auto_acts=False, fp_quant=True),
      dict(use_auto_acts=False, fp_quant=True, use_unstructured=True),
  )
  def test_schema_matches_expected(
      self,
      use_auto_acts,
      fp_quant,
      use_unstructured=False,
  ):
    # This tests that the schema of the configdict returned by 'base_config',
    # once all references are resolved, matches an expected schema. 'Schema'
    # here means the names and structure of fields at each level of the
    # configuration hierarchy. A value of 'None' in the expected schemas defined
    # below indicates a real configuration would have a concrete scalar value
    # there.

    if fp_quant:
      prec = {
          'fp_spec': {
              'exp_min': None,
              'exp_max': None,
              'sig_bits': None,
          },
          'is_scaled': None,
      }
    else:
      prec = None

    if use_auto_acts:
      quant_act_schema = {
          'bounds': {
              'initial_bound': None,
              'stddev_coeff': None,
              'absdev_coeff': None,
              'mix_coeff': None,
              'reset_stats': None,
              'ema_coeff': None,
              'use_cams': None,
              'exclude_zeros': None,
              'use_mean_of_max': None,
              'granularity': None
          },
          'input_distribution': None,
          'prec': prec,
          'half_shift': None,
      }
    else:
      quant_act_schema = {
          'bounds': None,
          'input_distribution': None,
          'prec': prec,
          'half_shift': None,
      }
    prune_rate = None if use_unstructured else [None, None]
    sparsity = {
        'type': None,
        'prune_rate': prune_rate,
        'smallest': None,
        'order': None,
        'absolute': None,
        'structure_decay': None,
        'sparse_ste': None
    }
    expected_top_level_schema = {
        'metadata': {
            'description': None,
            'hyper_str': None
        },
        'weight_decay': None,
        'activation_bound_update_freq': None,
        'activation_bound_start_step': None,
        'prec': prec,
        'half_shift': None,
        'weight_prec': prec,
        'weight_half_shift': None,
        'quant_type': None,
        'quant_act': quant_act_schema,
        'weight_quant_granularity': None,
        'sparsity': sparsity,  # Top level sparsity value, sets both weight and
        # activation sparsity,
        'weight_sparsity': sparsity,
        'act_sparsity': sparsity,
    }

    config = config_schema_utils.get_base_config(
        use_auto_acts=use_auto_acts,
        fp_quant=fp_quant,
        use_unstructured=use_unstructured,
    )
    # This round-trip conversion from JSON forces all references to resolve to
    # concrete values.
    config_reified = json.loads(config.to_json())

    # This test is not interested in checking the specific values of fields in
    # the configuration, but only that the schema of the hierarchies
    # are the same. Thus we all set the value of leaf nodes in the config to
    # 'None' before checking that the actual and expected configuration
    # structures are the same.
    def set_leaves_to_none(config):
      # We are at an intermediate node in the tree-structured input, which could
      # either be in the form of a dictionary or a list of other nodes in the
      # tree.
      if isinstance(config, dict):
        return {key: set_leaves_to_none(value) for key, value in config.items()}
      elif isinstance(config, list):
        return [set_leaves_to_none(value) for value in config]

      # We are at a leaf node in the tree-structured input.
      else:
        return None

    self.assertSameStructure(
        set_leaves_to_none(config_reified), expected_top_level_schema)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Injectable AQT ops.

Pax implements vanilla op wrappers in praxis/layers/base_ops.py
This file contains AQT variants of these.
"""

import aqt.jax.v2.aqt_dot_general as aqt
import aqt.jax.v2.config as aqt_config
import jax.numpy as jnp
from praxis import base_layer


class AqtEinsum(base_layer.BaseLayer):
  """Quantized Einsum class for model injection."""

  cfg: aqt_config.DotGeneral | None = None
  track_train_step: bool = False

  def setup(self) -> None:
    if self.track_train_step:
      self.create_variable(
          'train_step',
          base_layer.WeightHParams(
              shape=[],
              init=base_layer.WeightInit.Constant(0),
              dtype=jnp.int32,
          ),
          trainable=False,
      )

  def __call__(self, eqn, lhs, rhs):
    if self.track_train_step:
      if not self.do_eval:
        train_step = self.get_var('train_step')
        self.update_var('train_step', train_step + 1)
      # train_step starts from 0 and ends at exactly the total_train_step-1
      train_step = self.get_var('train_step')
    else:
      train_step = None
    cfg = self.cfg
    if cfg is not None:
      key = self.next_prng_key()
      cfg = aqt_config.set_context(cfg, key, train_step)
    dg = aqt.make_dot_general(cfg)
    # jnp.einsum is by default jitted, which makes the key storing in cfg
    # cross the jit boundary. We need to call a non-jitted jnp.einsum below
    return aqt.einsum(eqn, lhs, rhs, dg)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Basic functionalities for pruning neural networks implemented in jax."""

import dataclasses
import enum
import functools
import math
import typing
from typing import Tuple, Union

from absl import logging
from aqt.jax_legacy.jax.flax import struct as flax_struct
from flax import linen as nn
import jax
from jax import lax
import jax.numpy as jnp


dataclass = (
    flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass
)


# TODO(ayazdan): Create abstract class for sparsity configurations.
class SparseType(str, enum.Enum):
  """Pruning types dataclass."""
  STRUCTURED_NM = 'STRUCTURED_NM'
  UNSTRUCTURED = 'UNSTRUCTURED'


@dataclass
class SparseHParams:
  """Hyper parameters for sparsity.

  Attributes:
    type: Input array for which pruning mask is computed.
    prune_rate: Pruning rate.
    order: Apply pruning using this index order. Supported values are `C`, `R`.
      `C` and `R` indicate column-wise and row-wise ordering, respectively.
    absolute: If True, the absolute value of the values are used for sorting.
    smallest: If True, the smallest values in inputs are masked.
    structure_decay: If True, a decaying mechanism is applied on the structure.
    mask_decay_weight: If 0.0, no mask decay is applied. The mask value
      start with 1.0 and each time `num_update_sparsity` * `mask_decay_weight`
      is subtracted from 1.0. Due to overhead of jit, we limited the number of
      updates to `num_update_sparsity` to 16. After 16 iterations, we forcefully
      set `mask_decay_value` to zero. Mask decaying works for both structured
      and unstructured sparsity.
    sparse_ste: If True, a sparse-refined straight-through estimator (SR-STE)
      is applied, following the algorithm described in:
        https://arxiv.org/abs/2102.04010
    sparse_ste_weight: Denotes the relative weight for the sparse-refined term.
      As mentioned in the paper (https://arxiv.org/abs/2102.04010), the best
      default value is 0.0002 (lambda_w in the paper).
  """

  type: SparseType
  # Prune_rate data type varies with SparseHParams::type.
  # float: If SparseHParams::type is UNSTRUCTURED.
  # Tuple[int]: If SparseHParams::type is STRUCTURED_NM,  Tuple of type N, M.
  prune_rate: Union[None, float, Tuple[int, int]]
  order: str = 'R'
  absolute: bool = True
  smallest: bool = True
  structure_decay: bool = False
  mask_decay_weight: float = 0.0
  sparse_ste: bool = False
  sparse_ste_weight: float = 0.0002

  def __post_init__(self):
    if self.prune_rate is not None:
      if self.type == SparseType.STRUCTURED_NM:
        assert isinstance(self.prune_rate,
                          Tuple), ('prune rate should be either '
                                   'None for no pruning or a '
                                   'Tuple (N, M) for '
                                   'STRUCTURED_NM sparsity')
      elif self.type == SparseType.UNSTRUCTURED:
        assert isinstance(self.prune_rate,
                          float), ('prune rate should be either '
                                   'None for no pruning or float for '
                                   'UNSTRUCTURED sparsity')
      else:
        assert False, 'prune rate unknown!'

      assert self.mask_decay_weight >= 0.0, (
          'Invalid value for '
          f'{self.mask_decay_weight}. '
          '`mask_decay_weight` must be positive.')

      if self.sparse_ste:
        if self.mask_decay_weight != 0.0:
          raise ValueError('SR-STE only works with non-decaying mask.')
        if self.structure_decay:
          raise ValueError(
              'SR-STE only works with non-decaying sparse structure.')
        if self.type != SparseType.STRUCTURED_NM:
          raise ValueError('SR-STE only works with structured sparsity.')


@functools.partial(jax.custom_vjp, nondiff_argnums=(4, 5, 6))
def sr_ste(inputs: jnp.ndarray,
           mask: jnp.ndarray,
           update_mask: bool,
           apply_mask: bool,
           sparsity_hparams: SparseHParams,
           n_sparsity: int = 0,
           m_sparsity: int = 0):
  """Wrapper function for custom derivative rule for structured sparsity.

  Algorithm description: https://arxiv.org/abs/2102.04010

  The last three arguments are forced to be static to simplify
    the implementation.

  Args:
    inputs: Input array for which N:M pruning mask is computed.
    mask: The mask matrix which defines which elements to be pruned.
    update_mask: If True, the mask pattern gets updated.
    apply_mask: If True, the mask is applied to input.
    sparsity_hparams: The hyperparmeters related to sparsity.
    n_sparsity: Integer value for N in N:M sparsity.
    m_sparsity: Integer value for M in N:M sparsity.

  Returns:
    The updated input values after applying sparsity.
  """

  return sr_ste_fwd(
      inputs=inputs,
      mask=mask,
      update_mask=update_mask,
      apply_mask=apply_mask,
      sparsity_hparams=sparsity_hparams,
      n_sparsity=n_sparsity,
      m_sparsity=m_sparsity)[0]


@functools.partial(jax.jit, static_argnums=(4, 5, 6))
def sr_ste_fwd(inputs: jnp.ndarray,
               mask: jnp.ndarray,
               update_mask: bool,
               apply_mask: bool,
               sparsity_hparams: SparseHParams,
               n_sparsity: int = 0,
               m_sparsity: int = 0) -> jnp.ndarray:
  """Custom forward pass for structured sparsity."""
  # pylint:disable=g-long-lambda
  updated_mask = jax.lax.cond(
      update_mask, lambda: get_sparsity_mask(
          inputs, sparsity_hparams, n_sparsity, m_sparsity), lambda: mask)
  updated_inputs = jax.lax.cond(apply_mask,
                                lambda: jnp.multiply(updated_mask, inputs),
                                lambda: inputs)
  # pylint:enable=g-long-lambda
  return (updated_inputs, updated_mask,  # pytype: disable=bad-return-type  # jax-ndarray
          jnp.array(SparseHParams.sparse_ste_weight)), (
              inputs, updated_mask, jnp.array(SparseHParams.sparse_ste_weight))


def sr_ste_bwd(sparsity_hparams, n_sparsity, m_sparsity, res, g):
  """Implements custom gradient for backward pass.

  Args:
    sparsity_hparams: Non-diff arguments as defined in `sr_ste`.
    n_sparsity: Non-diff arguments as defined in `sr_ste`.
    m_sparsity: Non-diff arguments as defined in `sr_ste`.
    res: Residuals computed in sr_ste_fwd.
    g: Default calculated gradients.

  Returns:
    Gradients for differentiable inputs:
      - inputs
      - mask
      - update_mask
      - apply_mask
  """
  del sparsity_hparams, n_sparsity, m_sparsity
  inputs, updated_mask, ste_weight = res
  # g contains a list of gradients, one per output.
  # g1: updated_inputs
  g1, _, _ = g
  g1 = g1 + ste_weight * jnp.multiply(~updated_mask, inputs)
  return (g1, None, None, None)


sr_ste.defvjp(sr_ste_fwd, sr_ste_bwd)


class Sparsity(nn.Module):
  """Abstract class sparsity for applying sparsity."""

  sparsity_hparams: SparseHParams

  @nn.compact
  def __call__(self,
               inputs: jnp.ndarray,
               *,
               update_mask: bool,
               apply_mask: bool,
               num_update_sparsity: int = 0) -> jnp.ndarray:

    # TODO(shivaniagrawal): make a decision on creating/not creating mask for
    # when sparsity hparams is None itself.
    if (self.sparsity_hparams is None or
        self.sparsity_hparams.prune_rate is None):
      return inputs

    if self.sparsity_hparams.type == 'STRUCTURED_NM':
      n_sparsity = self.sparsity_hparams.prune_rate[0]
      m_sparsity = self.sparsity_hparams.prune_rate[1]
      if self.sparsity_hparams.structure_decay:
        if num_update_sparsity == 1:
          n_sparsity = n_sparsity - 1
        else:
          n_sparsity = int(
              math.ceil(n_sparsity / math.pow(2, num_update_sparsity)))
    else:
      logging.info('Unstructured sparsity does not support structure decaying.')
      n_sparsity = 0
      m_sparsity = 0

    # Due to overhead of jit, we limited the number of updates to
    # `num_update_sparsity` to 16. Once we reach to 16, we forcefully set
    # `mask_decay_value` to zero.
    # TODO(ayazdan): Support more than 16 decay.
    mask_decay_value = 1.0
    if self.sparsity_hparams.mask_decay_weight != 0.0:
      if num_update_sparsity < 16:
        mask_decay_value = max(
            mask_decay_value -
            (num_update_sparsity * self.sparsity_hparams.mask_decay_weight),
            0.0)
      else:
        mask_decay_value = 0.0
    mask = self.variable('sparsity', 'mask', jnp.ones, inputs.shape, jnp.bool_)

    if self.sparsity_hparams.sparse_ste:
      updated_inputs, updated_mask, _ = sr_ste(
          inputs=inputs,
          mask=mask.value,
          update_mask=update_mask,
          apply_mask=apply_mask,
          sparsity_hparams=self.sparsity_hparams,
          n_sparsity=n_sparsity,
          m_sparsity=m_sparsity)
      if update_mask and self.has_variable('sparsity', 'mask'):
        mask.value = updated_mask
      return updated_inputs
    else:
      if update_mask and self.has_variable('sparsity', 'mask'):
        mask.value = get_sparsity_mask(inputs, self.sparsity_hparams,
                                       n_sparsity, m_sparsity)

      if apply_mask and self.has_variable('sparsity', 'mask'):
        if self.sparsity_hparams.mask_decay_weight != 0.0:
          return jnp.multiply(~mask.value * mask_decay_value + mask.value,
                              inputs)
        else:
          return jnp.where(mask.value, inputs,
                           jnp.zeros(inputs.shape, inputs.dtype))
      return inputs


def apply_sparsity(
    inputs: jnp.ndarray,
    sparsity_hparams: SparseHParams,
    n_sparsity: int,
    m_sparsity: int,
) -> jnp.ndarray:
  """Returns sparsified inputs based on sparsity hparams."""
  mask = get_sparsity_mask(inputs, sparsity_hparams, n_sparsity, m_sparsity)
  return jnp.where(mask.value, inputs, jnp.zeros(inputs.shape, inputs.dtype))  # pytype: disable=attribute-error  # jax-ndarray


def get_sparsity_mask(
    inputs: jnp.ndarray,
    sparsity_hparams: SparseHParams,
    n_sparsity: int = 0,
    m_sparsity: int = 0,
) -> jnp.ndarray:
  """Returns sparsified inputs based on sparsity hparams."""
  if sparsity_hparams is None or sparsity_hparams.prune_rate is None:
    return jnp.ones(inputs.shape, dtype=bool)
  prune_rate = sparsity_hparams.prune_rate
  if sparsity_hparams.type == 'STRUCTURED_NM':
    assert isinstance(
        prune_rate, Tuple), 'prune rate must be tuple for structured sparsity.'
    assert prune_rate[0] <= prune_rate[1], (
        'prune_rate[0] must be lower than prune_rate[1] for N:M'
        f' ({prune_rate[0]}:{prune_rate[1]}) sparsity.'
    )
    return get_pruning_n_m_mask(
        inputs,
        n=n_sparsity,
        m=m_sparsity,
        order=sparsity_hparams.order,
        absolute=sparsity_hparams.absolute,
        smallest=sparsity_hparams.smallest)
  elif sparsity_hparams.type == 'UNSTRUCTURED':
    assert (
        isinstance(prune_rate, float) and prune_rate < 1
    ), f'sparsity ratio can not be > 1, provided prune_rate {prune_rate}.'
    return get_pruning_unstruct_mask(
        inputs, prune_rate=prune_rate, smallest=sparsity_hparams.smallest)
  else:
    raise ValueError(f'invalid sparsity type {sparsity_hparams.type}!')


# TODO(ayazdan): Support arrays with length not divisible by `m`.
def prune_inputs_n_m(inputs: jnp.ndarray,
                     *,
                     n: int,
                     m: int,
                     order: str = 'R',
                     offset: int = 0,
                     absolute: bool = True,
                     smallest: bool = True) -> jnp.ndarray:
  """Returns pruned array with N:M (structured) pruning.

  N:M pruning makes at most N values non-zero in each block of M consecutive
  values.

  Args:
    inputs: Input array for which N:M pruning mask is computed.
    n: Maximum number of non-zero values in each block.
    m: Number of values in each block.
    order: Apply pruning using this index order. Supported values are `C`, `R`.
    `C` -> Column-wise pruning.
    `R` -> Row-wise pruning.
    offset: Indicates the offset between the group of M elements on which
      N:M sparsity is applied. The default is `0` (narrowly-separate),
      indicating that `M` elements are selected from adjacent values in
      the input matrix. Generally, because of the XLA layout
      (lanes 128/sublanes 8), another value for offset would be 128
      (widely-separated).
      If offset > 0, we only support scenarios where the input array size is
      equal to (offset * m).
    absolute: If True, the absolute value of the values are used for sorting.
    smallest: If True, the smallest values in inputs are masked.

  Returns:
    An array with the same shape as inputs pruned with N:M strategy.
  """
  mask = get_pruning_n_m_mask(
      inputs,
      n,
      m,
      order=order,
      offset=offset,
      absolute=absolute,
      smallest=smallest)
  return jnp.where(mask, inputs, jnp.zeros(inputs.shape, inputs.dtype))


# TODO(ayazdan): Add supports for magnitude pruning.
# TODO(ayazdan): Add supports for column-wise pruning.
# TODO(ayazdan): Add for other form of N:M pruning.
# TODO(ayazdan): Add parameter `r` for decaying weights.
def prune_2_4(inputs: jnp.ndarray) -> jnp.ndarray:
  """Returns pruned array w/ 2:4 (structured) pruning (w/ sorting networks).

  Args:
    inputs: Input array for which N:M pruning mask is computed.

  Returns:
    An array with the same shape as inputs pruned with 2:4 strategy.
  """
  def _cswap(a, b, cond):
    """Conditional swap of two vector."""
    return jnp.where(cond, a, b), jnp.where(cond, b, a)

  def _prune(a, n, r):
    """Prunes the first n elements with a decaying factor of (1-r).

    Args:
      a: Input array on which pruning is applied.
      n: The first number of elements on which mask is applied.
      r: The decaying factor for sparsification.

    Returns:
      A sparsified array with the same shape as `a`.
    """
    for i in range(n):
      a[i] = a[i] * (1 - r)
    return a

  def _preprocessing(xin, m=4):
    """Reorganizes the input vector for sorting.

    Example:
      Input:
        (0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (0, 7)
      Output:
        (0, 0) - (0, 4)
        (0, 1) - (0, 5)
        (0, 2) - (0, 6)
        (0, 3) - (0, 7)

    Args:
      xin: Input array on which the reorganizatin occurs.
      m: Numner of adjacent elements to group.

    Returns:
      A reorganized array with the same total size as input.
      In addition, it returns `start` and `end` indices of each dimension.
      Finally, it returns `red_dim` which indicates the last dimension of input
      array divided by group size (m).
    """
    red_dim = int(jnp.shape(xin)[-1] / m)
    start = [0] * len(jnp.shape(xin))
    end = list(jnp.shape(xin))
    strides = [1] * (len(jnp.shape(xin)) - 1) + [m]
    xs = []
    for i in range(m):
      start_i = start
      start_i[-1] = i
      xs.append(
          jnp.reshape(lax.slice(xin, start_i, end, strides), [-1, red_dim]))
    return xs, start, end, red_dim

  def _postprocessing(xin, start, end, red_dim, shape_in):
    """Reorganizes the input vector back to its original shape.

    Example:
      Input:
        (0, 0) - (0, 4)
        (0, 1) - (0, 5)
        (0, 2) - (0, 6)
        (0, 3) - (0, 7)
      Output:
        (0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (0, 7)

    Args:
      xin: Input array on which the reorganizatin occurs.
      start: The starting indices of each dimension.
      end: The ending indices of each dimension.
      red_dim: Indicates the last dimension of input array divided
        by group size (m).
      shape_in: The original shape of the input array.

    Returns:
      A reorganized array back to its original shape after sparsification.
    """

    filtered = jax.numpy.reshape(xin, shape_in)
    slices = []
    strides = [1] * (len(shape_in) - 1) + [red_dim]
    for i in range(red_dim):
      start_i = start
      start_i[-1] = i
      slices.append(lax.slice(filtered, start_i, end, strides))
    return jax.numpy.reshape(jax.numpy.stack(slices, axis=-2), shape_in)

  # pre-processing: preprare the input array for sorting network.
  xs, start, end, red_dim = _preprocessing(inputs)
  comparators = [(0, 1), (2, 3), (0, 2), (1, 3), (1, 2)]
  comparator_results = []
  # forward pass through sorting network.
  for x, y in comparators:
    l = jnp.less(xs[x], xs[y])
    comparator_results.append(l)
    xs[x], xs[y] = _cswap(xs[x], xs[y], l)
  # apply pruning.
  pruned = _prune(a=xs, n=2, r=1.0)
  # backward pass through sorting network.
  for xy, cond in reversed(list(zip((comparators), comparator_results))):
    x, y = xy
    pruned[x], pruned[y] = _cswap(pruned[x], pruned[y], cond)
  xs_final = jnp.array(jnp.concatenate(pruned, axis=1), dtype=xs[0].dtype)
  # post-processing: reshape the array to its original form.
  return _postprocessing(xs_final, start, end, red_dim, jax.numpy.shape(inputs))


def get_pruning_n_m_mask(inputs: jnp.ndarray,
                         n: int,
                         m: int,
                         *,
                         order: str = 'R',
                         offset: int = 0,
                         absolute: bool = True,
                         smallest: bool = True) -> jnp.ndarray:
  """Returns a mask for N:M (structured) pruning.

  N:M pruning makes at most N values non-zero in each block of M consecutive
  values.

  Args:
    inputs: Input array for which N:M pruning mask is computed.
    n: Maximum number of non-zero values in each block.
    m: Number of values in each block.
    order: Apply pruning using this index order. Supported values are `C`, `R`.
      `C` and `R` indicate column-wise and row-wise masking, respectively.
      Default is `R` indicating to applying N:M sparsity across rows of
      the input matrix.
    offset: Indicates the offset between the group of M elements on which
      N:M sparsity is applied. The default is `0` (narrowly-separate),
      indicating that `M` elements are selected from adjacent values in
      the input matrix. Generally, because of the XLA layout
      (lanes 128/sublanes 8), another value for offset would be 128
      (widely-separated).
      If offset > 0, we only support scenarios where the input array size is
      equal to (offset * m).
    absolute: If True, the absolute value of the values are used for sorting.
    smallest: If True, the smallest values in inputs are masked.

  Returns:
    A mask that indicates the pruning locations (`0`: no pruning, `1`: pruned).
  """
  if n > m:
    raise ValueError(f'N must be lower than M for N:M ({n}:{m}) sparsity.')
  if order not in ['C', 'R']:
    raise ValueError(f'Index order {order} not supported.')
  if offset < 0:
    raise ValueError('Offset value must be positive. '
                     f'You provided {offset}.')
  if offset != 0 and offset != 128:
    logging.warning(
        'Value %d may not be best optimized for the memory layout. '
        'We suggest a value of `0` or `128` for offset.', offset
    )
  length = jnp.size(inputs)
  # TODO(b/228458062): support m which is not a factor of inputs size.
  if length % m != 0:
    raise ValueError(
        f'inputs size must be divisible by m, provided {length} and {m}')
  group = int(length / m)
  inputs = jnp.abs(inputs) if absolute else inputs
  if order == 'R':
    if offset == 0:
      inputs_temp = inputs.reshape(group, m, order='C')
    else:
      if offset * m != length:
        raise ValueError('When offset > 0, we only support an array size '
                         '(length) equal to (offset * m). '
                         f'Provided offset = {offset}, m = {m}, '
                         f'length = {length}.')
      inputs_temp = jnp.transpose(inputs.reshape(m, offset, order='C'))
  else:
    inputs_temp = jnp.einsum('...ij->...ji', inputs).reshape(
        group, m, order='C')
  mask = jnp.ones(inputs_temp.shape, dtype=bool)
  # Extract the smallest elements and forcefully make them zero.
  _, top_k_indices = jax.lax.top_k(
      -inputs_temp, k=m - n) if smallest else jax.lax.top_k(
          inputs_temp, k=m - n)
  mask = jax.vmap(lambda x, i: x.at[i].set(False))(mask, top_k_indices)
  if order == 'R':
    if offset == 0:
      return mask.reshape(inputs.shape, order='C')
    else:
      return jnp.transpose(mask).reshape(inputs.shape, order='C')
  else:
    if len(inputs.shape) > 2:
      return jnp.einsum('...ij->...ji', mask.reshape(inputs.shape, order='F'))
    else:
      return jnp.einsum('ij->ji', mask).reshape(inputs.shape, order='F')


def get_pruning_unstruct_mask(inputs: jnp.ndarray,
                              *,
                              prune_rate: float = 0.1,
                              smallest: bool = True) -> jnp.ndarray:
  """Returns a mask for pruning according to the prune rate.

  Args:
    inputs: Input array for which pruning mask is computed.
    prune_rate: Pruning rate. The ratio of the elements that are pruned. 0
      meaning no pruning. Defaults to 0.1.
    smallest: If True, the smallest values in inputs are masked.

  Returns:
    A mask that indicates the pruning locations.
  """
  inputs = -jnp.abs(inputs) if smallest else jnp.abs(inputs)
  mask = jnp.ones(inputs.shape, dtype=bool)
  k = int(inputs.size * prune_rate)
  _, idxs = jax.lax.top_k(inputs.reshape(-1), k)
  return mask.reshape(-1).at[idxs].set(False).reshape(inputs.shape)


def prune_inputs_unstruct(inputs: jnp.ndarray,
                          *,
                          prune_rate: float = 0.1,
                          smallest: bool = True) -> jnp.ndarray:
  """Returns unstructured pruning for inputs according to the prune rate.

  Args:
    inputs: Input array which is being pruned.
    prune_rate: Pruning rate. The ratio of the elements that are pruned. 0
      meaning no pruning. Defaults to 0.1.
    smallest: If True, the smallest values in inputs are masked.

  Returns:
    Pruned input.
  """
  mask = get_pruning_unstruct_mask(
      inputs, prune_rate=prune_rate, smallest=smallest)
  return jnp.where(mask, inputs, jnp.zeros(inputs.shape, inputs.dtype))



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Resnet50 quantized model.

# pylint: disable=line-too-long

Results at the last batch from TB:
- Eval accuracy: 0.7271
- Eval loss: 1.077
- Train accuracy: 0.7365
- Training loss: 1.102
- Tensorboard link: http://shortn/_wpbXEvjgmx
- Xmanager link: http://shortn/_KAzZi6QzHI
- CNS link: http://shortn/_kamgdDcOKe
- Metrics from reporting tool: http://shortn/_Ey7SVZ1jeY
- Bash file for launching: ./resnet50_w1_a4_auto.sh
"""

from aqt.jax_legacy.jax.imagenet.configs import base_config


def get_config(quant_target=base_config.QuantTarget.WEIGHTS_AND_AUTO_ACT):
  """Gets Resnet50 config for 1 bit weights and 4 bits auto activation quantization.

  conv_init and last dense layer not quantized as these are the most
  sensitive layers in the model.

  Args:
   quant_target: quantization target, of type QuantTarget.

  Returns:
   ConfigDict instance.
  """
  config = base_config.get_config(
      imagenet_type=base_config.ImagenetType.RESNET50,
      quant_target=quant_target)
  config.weight_prec = 1
  config.quant_act.prec = 4
  config.half_shift = True

  # The first conv layer and the dense layer still use 8-bit weights and acts
  config.model_hparams.conv_init.weight_prec = 8
  config.model_hparams.conv_init.quant_act.prec = 8

  config.model_hparams.dense_layer.weight_prec = 8
  config.model_hparams.dense_layer.quant_act.prec = 8

  config.metadata.hyper_str = "resnet50_w1_a4_auto"

  return config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pandas dataframe wrapper functions for frequently used operations."""
import functools
import re
from typing import Any, List, Optional, Sequence, Tuple, Union

import pandas as pd

Regex = str
NewStr = str
ColumnName = str
ColumnValue = Any
OrderAscending = bool
SingleColumnFilter = Tuple[ColumnName, List[ColumnValue]]
SingleColumnRegex = Tuple[ColumnName, Regex]
SingleColumnRegexReplace = Tuple[ColumnName, Regex, Regex]
ColumnNameRegexReplace = Tuple[Regex, Regex]
SortBy = Sequence[Tuple[ColumnName, OrderAscending]]


def select_rows_by_column_values(df: pd.DataFrame, column_name: str,
                                 values: Union[Any, List[Any]]) -> pd.DataFrame:
  """Selects rows where column values are in specified list.

  Args:
    df: A pandas dataframe.
    column_name: Name of the column to filter the rows by.
    values: Single value or list of values to filter by.

  Returns:
    Filtered dataframe.

  """
  if not isinstance(values, List):
    values = [values]
  return df[df[column_name].isin(values)]


def select_rows_by_regex(df: pd.DataFrame, column_name: str,
                         regex_str: str) -> pd.DataFrame:
  """Returns pandas dataframe rows matching regex in a specific column.

  Returns a dataframe with a subset of rows where the value in the column
  contains the regex.

  Args:
    df: A pandas dataframe.
    column_name: Name of the column to select the rows by. Has to contain string
      values only.
    regex_str: a string or regex string to match column values against.

  Returns:
    Filtered dataframe.

  """
  return df[df[column_name].str.contains(regex_str)]


def drop_rows_by_column_values(df: pd.DataFrame, column_name: str,
                               values: List[Any]) -> pd.DataFrame:
  """Drop rows where their column values are specified list.

  Args:
    df: A pandas dataframe.
    column_name: Name of the column to filter the rows by.
    values: List of values to filter by.

  Returns:
    Filtered dataframe.

  """
  return df[~df[column_name].isin(values)]


def drop_rows_by_regex(df: pd.DataFrame, column_name: str,
                       regex_str: str) -> pd.DataFrame:
  """Returns pandas dataframe rows matching regex in column dropped.

  Args:
    df: A pandas dataframe.
    column_name: Name of the column to filter the rows by. Has to contain string
      values only.
    regex_str: a string or regex string to match column values against.

  Returns:
    Filtered dataframe.

  """
  return df[~df[column_name].str.contains(regex_str)]


def filter_columns(df: pd.DataFrame, columns_to_keep=List[str]) -> pd.DataFrame:
  """Returns dataframe with subset of columns."""
  columns_to_keep = [c for c in columns_to_keep if c in list(df.columns)]
  return df[columns_to_keep]


def filter_columns_by_regex(df: pd.DataFrame,
                            column_regex: str) -> pd.DataFrame:
  """Returns dataframe with subset of columns matching regex."""
  return df.filter(regex=column_regex)


def drop_columns(df: pd.DataFrame, columns_to_drop=List[str]) -> pd.DataFrame:
  """Returns dataframe with specified columns dropped."""
  columns_to_drop = [c for c in columns_to_drop if c in list(df.columns)]
  return df.drop(columns=columns_to_drop)


def drop_columns_by_regex(df: pd.DataFrame, column_regex: str) -> pd.DataFrame:
  """Returns dataframe with columns matching the regex dropped."""
  return df[df.columns.drop(list(df.filter(regex=column_regex)))]


def group_by_with_aggregation(
    df: pd.DataFrame,
    by: Union[str, List[str]],
    agg_column_names: List[str],
    aggregators: Tuple[str, ...] = ('mean', 'std')
) -> pd.DataFrame:
  """Group by with aggregation.

  Args:
    df: Dataframe to perform group by on.
    by: Which column to group by.
    agg_column_names: Which columns to aggregate over. Have to be columns
      containing numeric values.
    aggregators: Which aggregation functions to apply.

  Returns:
    Multi-level dataframe where each row corresponds to a group and contains
    aggregated values for specified agg_column_names and aggregators.

  """
  agg_dict = {}
  for col in agg_column_names:
    agg_dict[col] = aggregators

  return df.groupby(by, as_index=False).agg(agg_dict)


def rename_values_in_column(df: pd.DataFrame, column_name: str, pattern: Regex,
                            repl: Regex) -> pd.DataFrame:
  """Renames row values of dataframe by regex replacement.

  Args:
    df: A pandas dataframe.
    column_name: In which column to rename values. Has to be string column.
    pattern: Regex pattern to replace.
    repl: Replacement string to insert where old string was.

  Returns:
    Dataframe with renamed columns.
  """
  df[column_name] = df[column_name].apply(
      lambda x: re.sub(pattern=pattern, repl=repl, string=x))
  return df


def rename_column_headers(df: pd.DataFrame, pattern: Regex,
                          repl: Regex) -> pd.DataFrame:
  """Renames column headers of dataframe by regex replacement.

  Args:
    df: A pandas dataframe.
    pattern: Regex pattern to replace.
    repl: Replacement string to insert where old string was.

  Returns:
    Dataframe with renamed columns.
  """

  def _regex_replace_fn(x: str, pattern: Regex, repl: Regex) -> str:
    return re.sub(pattern=pattern, repl=repl, string=x)

  return df.rename(
      columns=functools.partial(_regex_replace_fn, pattern=pattern, repl=repl))


def apply_filter_drop_rename_operations(
    df: pd.DataFrame,
    row_filter_args: Optional[Sequence[SingleColumnFilter]] = None,
    row_regex_filter_args: Optional[Sequence[SingleColumnRegex]] = None,
    rename_row_value_args: Optional[Sequence[SingleColumnRegexReplace]] = None,
    drop_columns_by_regex_args: Optional[Sequence[Regex]] = None,
    rename_column_name_args: Optional[Sequence[ColumnNameRegexReplace]] = None,
    sort_by_args: Optional[SortBy] = None) -> pd.DataFrame:
  r"""Function to apply series of filter, drop and rename operations.

  Tthe user can specify how to filter the dataframe rows, which
  columns to drop, and how to rename columns. For more customized pandas
  operations, please use report_utils.load_all_reports() to load all reports
  into a dataframe.

  Usage example:
    df = pandas_utils.apply_filters_and_renaming(
    df,
    row_filter_args=
      [('xid', [18925472, 18925219, 18925394])],
    row_regex_filter_args=
      [('model_dir', '.*4bit.*')],
    rename_row_value_args=
      [('experiment_name', r'leaderboard_full_model_(.*)_wanglisa-.*',
      r'\g<1>')],
    drop_columns_by_regex_args=['.*unsmoothed.*'],
    rename_column_name_args=[('_translate--de-en:test', '')],
    sort_by_args=([('eval/loss', True), ('eval/bleu', False)]),
  )

  Args:
    df: Dataframe to run filter and rename operations on.
    row_filter_args: List of row filters to be applied sequentially. Each row
      filter is a tuple of (column_name, filter_list). Each row filter is
      equivalent to a SQL statement of the form SELECT * WHERE column_name IN
      filter_list.
    row_regex_filter_args: Each row filter is a tuple of (column_name, regex).
      Will select the rows where column values match the regex.
    rename_row_value_args: Arguments for replacing values in columns. List of
      tuples (column_name, old_string, new_string).
    drop_columns_by_regex_args: Drop columns matching the provided regex.
    rename_column_name_args: Arguments to rename column headers. List of tuples
      (old_string (can be regex), new_string).
    sort_by_args: Arguments for df.sort_values(), List of tuples of (by,
      ascending), where `by` is a column sort by, and `ascending` is a bool
      indicating whether the column should be sorted in ascending order. First
      tuple will be used as primary sorting axis, etc. See documentation
      pandas.DataFrame.sort_values() for more details.

  Returns:
    A pandas dataframe.

  """
  for column_name, value in row_filter_args or []:
    df = select_rows_by_column_values(df, column_name=column_name, values=value)

  for column_name, regex in row_regex_filter_args or []:
    df = select_rows_by_regex(df, column_name=column_name, regex_str=regex)

  for column_name, pattern, repl in rename_row_value_args or []:
    df = rename_values_in_column(
        df, column_name=column_name, pattern=pattern, repl=repl)

  for regex in drop_columns_by_regex_args or []:
    df = drop_columns_by_regex(df, column_regex=regex)

  for pattern, repl in rename_column_name_args or []:
    df = rename_column_headers(df, pattern=pattern, repl=repl)

  if sort_by_args is not None:
    by = [tup[0] for tup in sort_by_args]
    ascending = [tup[1] for tup in sort_by_args]
    df = df.sort_values(by=by, ascending=ascending)

  df = df.reset_index(drop=True)
  return df


def boxplot_with_group_by_sorted_by_median(df: pd.DataFrame,
                                           column_to_plot: str,
                                           by: Union[str, List[str]]):
  """Returns a box plot for dataframe column, after group by operation.

  Args:
    df: Dataframe to plot.
    column_to_plot: Which column to plot values of.
    by: Which column(s) to use to group by. Can be any valid input to
      pandas.DataFrame.groupby().

  Returns:
    A matplotlib boxplot of type matplotlib.axes._subplots.AxesSubplot.

  """
  grouped_df = pd.DataFrame(
      {group: values[column_to_plot] for group, values in df.groupby(by)})
  medians = grouped_df.median().sort_values()  # pytype: disable=attribute-error  # pandas-15-upgrade
  return grouped_df[medians.index].boxplot(rot=0, return_type='axes')



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration dataclasses."""

from typing import Literal
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import calibration
from aqt.jax.v2 import utils
from aqt.jax.v2.numerics import no_numerics
from aqt.jax.v2.numerics import numerics
import jax
import jax.numpy as jnp


AbstractAqtNumerics = numerics.AqtNumerics
AbstractAqtCalibration = calibration.Calibration


@utils.flax_slots_dataclass
class Context:
  key: jax.Array | None = utils.dynamic_field()
  train_step: int | None = utils.dynamic_field()


@utils.flax_slots_dataclass
class Quantizer:
  """Configuration of quantization of one tensor."""

  numerics: AbstractAqtNumerics = utils.static_field()
  calib_shared_axes: list[int] | Literal["per_tensor"] | None = (
      utils.static_field()
  )
  scale_stop_grad: bool = utils.static_field()
  # noise+clip+round
  # We apply gradient of clip_and_round in bwd pass.
  calibration: AbstractAqtCalibration = utils.static_field()
  # Round up the calibration to power of 2 (po2).
  po2_scale: bool = utils.static_field()
  # TODO(yichizh): Factor out auxilliary dataclasses into a separate file.
  context: Context

  # TODO(yichizh): Need to add type annotation back to cfg.
  def quant(
      self,
      x,
      *,
      calibration_axes,
  ) -> tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn]:
    """The core quantizing function."""
    if isinstance(self.numerics, no_numerics.NoNumerics):
      qt = aqt_tensor.QTensor(
          qvalue=x, scale=[], scale_t=[], dequant_dtype=x.dtype
      )
      return qt, None

    qt = self.calculate_scale(x, calibration_axes=calibration_axes)
    qt, quant_grad = self.calculate_qvalue(x, qt)
    return qt, quant_grad

  def calculate_scale(
      self,
      x,
      *,
      calibration_axes,
  ) -> aqt_tensor.QTensor:
    """Create incomplete QTensor with only quantization parameters."""
    dequant_dtype = x.dtype
    # TODO(lew): We should cast earlier. xhs_q should be in cfg.xhs.dtype
    # TODO(lew): After we implement optimization to not double-quantize,
    #   what would happen if we pass fq value (xhs_q2) in residual?
    if self.calib_shared_axes == "per_tensor":
      shared_axes = list(range(x.ndim))
    else:
      shared_axes = self.calib_shared_axes or calibration_axes
    bound = self.calibration.get_bound(x, shared_axes)
    abs_max_mapped_to = self.numerics.abs_val_mapped_to()
    scale = bound / abs_max_mapped_to

    if self.po2_scale:
      # With floor the biggest value (we are using jnp.max) is in the range of
      # clipping and therefore have a correct gradinet.
      scale = 2 ** jnp.floor(jnp.log2(jax.lax.reciprocal(scale)))
      scale = jax.lax.reciprocal(scale)
    if self.scale_stop_grad:
      # TODO(lew): Does not matter in DG, because we are using custom gradient.
      #   We should take that into account somehow.
      scale = jax.lax.stop_gradient(scale)

    qt = aqt_tensor.QTensor(
        qvalue=None,
        scale=[scale],
        scale_t=None,
        dequant_dtype=dequant_dtype,
    )
    return qt

  def calculate_qvalue(
      self,
      x,
      qt: aqt_tensor.QTensor
  ) -> tuple[aqt_tensor.QTensor, aqt_tensor.GradientFn]:
    """Uses the quantization parameters in qt to quantize x."""
    qt = qt.quant(x)

    # TODO(lew): A logical thing would be if this call was part of
    # QTensor.quant.
    x_q, res = self.numerics.vjp_fwd(qt.qvalue, self.context)
    quant_grad = jax.tree_util.Partial(self.numerics.vjp_bwd, res)

    qt = qt.replace(qvalue=x_q)
    return qt, quant_grad


def make_fake_quant(cfg, calibration_axes=None):
  def fake_quant(x):
    x_q, _ = cfg.quantizer.quant(x, calibration_axes=calibration_axes)
    return x_q.dequant()

  return fake_quant



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Flags shared by train.py and xm_launch.py.

These flags are shared by the traiing executable (train.py)
and the script that launches XMananger experiments (xm_launch.py).
"""

from absl import flags

FLAGS = flags.FLAGS

flags.DEFINE_string(
    'data_dir', default=None, help='Tensorflow datasets directory.')

flags.DEFINE_integer(
    'stats_serialization_freq', 200,
    'The number of steps in between serializing activation statistics.')

flags.DEFINE_string(
    'vocab_path',
    default=None,
    help='Path to load or store sentencepiece vocab file.')

flags.DEFINE_string(
    'dataset_name',
    default='wmt17_translate/de-en',
    help='Name of TFDS translation dataset to use.')

flags.DEFINE_string(
    'eval_dataset_name',
    default='wmt14_translate/de-en:test',
    help='Optional name of TFDS translation dataset to use for evaluation.')

flags.DEFINE_bool(
    'run_train_eval',
    default=False,
    help='Whether to run eval on training data/ used for eval debuggability.')

# NOTE: if changing this to DEFINE_list, verify if the behavior is as expected.
flags.DEFINE_multi_string(
    'additional_eval_datasets',
    default=[],
    # [ 'wmt15_translate/de-en:test', 'wmt16_translate/de-en:test',
    #  'wmt17_translate/de-en:test'],
    help='Names for additional eval datasets.')

flags.DEFINE_bool(
    'reverse_translation',
    default=True,
    help='Reverse the direction of translation.')

flags.DEFINE_integer(
    'eval_batch_size',
    default=256,
    help='Per host eval batch size for training.')

flags.DEFINE_integer(
    'eval_frequency',
    default=1087,
    help='Frequency of eval during training, e.g. every 1000 steps.')

flags.DEFINE_integer(
    'num_eval_steps', default=20, help='Number of evaluation steps.')

flags.DEFINE_integer(
    'max_target_length',
    default=256,
    help='Maximum length of training examples.')

flags.DEFINE_integer(
    'max_eval_target_length',
    default=97,
    help='Maximum length of eval examples.')

flags.DEFINE_integer(
    'max_predict_length',
    default=147,
    help='Maximum length for predicted tokens.')

flags.DEFINE_bool(
    'save_checkpoints',
    default=True,
    help='Whether to save model checkpoints for debugging.')

flags.DEFINE_bool(
    'save_minimum_loss_checkpoint',
    default=True,
    help='Whether to save a model checkpoint when eval loss is miniminal.')

flags.DEFINE_bool(
    'estimate_compute_and_memory_cost',
    default=False,
    help='Whether to estimate compute and memory cost.')

flags.DEFINE_bool(
    'restore_checkpoints',
    default=True,
    help='Whether to restore from existing model checkpoints.')

flags.DEFINE_string(
    'restore_checkpoint_model_dir',
    default=None,
    help='Directory to load checkpoint from when training first begins. Note '
    'that if a checkpoint already exists in "model_dir", that checkpoint is '
    'loaded instead under the assumption that this indicates the training run '
    'was preempted on Borg and is now being restarted.')

flags.DEFINE_integer(
    'checkpoint_freq',
    default=10000,
    help='Number of training steps between checkpoints.')

flags.DEFINE_bool(
    'use_bfloat16',
    default=True,
    help=('Use bfloat16 mixed precision training instead of float32.'))

flags.DEFINE_bool(
    'compute_train_metrics',
    default=True,
    help='Whether to compute the metrics during training.')

flags.DEFINE_string(
    'output_hlo_filename',
    default=None,
    help=(
        'Output model HLO to filename in FLAGS.model_dir. Both HLO text'
        ' (filename.txt) and protobuf (filename.pb) are emitted. If None, or'
        ' empty string, no HLO files will be written.'
    ),
)

flags.DEFINE_string(
    'output_beam_hlo_filename',
    default=None,
    help='Output model HLO, including the decoding beam search, to filename in'
    'FLAGS.model_dir. Both HLO text (filename.txt) and protobuf (filename.pb)'
    ' are emitted.')

flags.DEFINE_integer(
    'sparsity_start_step',
    default=-1,
    help=(
        'First step to compute sparsity mask at, defaults to -1 in which case '
        'we do not compute sparsity mask.'))

flags.DEFINE_integer(
    'sparsity_update_freq',
    default=0,
    help=(
        'How often (number of steps )to update sparsity mask, defaults to 0 in '
        'which case we do not update again.'
    ),
)

flags.DEFINE_bool(
    'visualize_acts_bound',
    default=False,
    help=(
        'Whether to visualize activations bounds for auto-clip in Tensorboard.'
        ' The bounds appear as scalar and will be named as "GetBounds_0/bounds"'
        ' prefixed with the all the parents module name.'))

flags.DEFINE_bool(
    'collect_acts_stats',
    default=False,
    help=('Whether to collect activation statistics and visualize them '
          'distributions in Tensorboard.'))

flags.DEFINE_integer(
    'state_dict_summary_freq',
    default=200,
    help=(
        'Number of training steps between state dict summaries reported to '
        'Tensorboard. Relevant to --visualize_acts_bound and --collect_acts_stats.'
    ))




# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common utility functions used for both TF and Jax."""

from typing import Optional, Sequence, Union

from aqt.common import aqt_config
from aqt.common import emulation_utils

_get_max_number_float = emulation_utils.get_max_number_from_mantissa_and_max_exp


def check_shapes_conformal(actual: Sequence[Optional[int]],
                           expected: Sequence[Optional[int]]):
  """Checks that actual shape conforms to expected.

  The `actual` shape conforms to the `expected` shape if it is equal in
  rank and has identical sizes in all non-None dimensions of `expected`.

  Note this is not a symmetric relation, like tf.ensure_shape. A tensor
  of completely-unknown size meets this `tf.ensure_shape` and
  `tf.TensorShape.is_compatible_with` relation, but is not conformal
  to any tensor with a known dimension.

  Args:
    actual: a list of ints for a tensor shape, with None for unknown dim.
    expected: the requirements for the shape, with None for no requirement.

  Raises:
    ValueError if the actual shape does not conform to expectations.
  """
  expected_shape = [
      ax if ax is not None else given for ax, given in zip(actual, expected)
  ]
  compatible_rank = len(actual) == len(expected)
  compatible_size = list(actual) == expected_shape
  if not compatible_rank or not compatible_size:
    raise ValueError(f'actual shape ({actual}) must be compatible with '
                     f'expected ({expected})')


def _get_clip_bound_int(config: aqt_config.IntQuantConfig):
  """Returns the clip bound when using integer values."""
  assert config.bits <= 23, 'Too many bits, float32 has less precision.'
  bucket_count = 2.0**config.bits
  if config.preserve_zero:
    bucket_count -= 1.0
  return bucket_count / 2.0


def get_clip_bound(
    config: Union[aqt_config.IntQuantConfig, aqt_config.SmallFloatConfig],
) -> float:
  """Returns the clip bound for IntQuantConfig or SmallFloatConfig."""
  config.validate()
  if isinstance(config, aqt_config.IntQuantConfig):
    return _get_clip_bound_int(config)
  elif isinstance(config, aqt_config.SmallFloatConfig):
    return _get_max_number_float(config.mantissa_bits, config.max_exp)
  else:
    raise ValueError(
        '_get_clip_bound called without quantization or emulation.')


def safe_clip_bound(config: aqt_config.IntQuantConfig) -> float:
  """Returns clip-safe clip bound.

  `get_clip_bound()` can be used for rescaling, but not clipping. This is
  because round-to-nearest-even (RTNE) behavior may round the clipped value out
  of range.  For instance, an 8-bit signed integer has a max value of 127; a
  clip of 127.5 may be appropriate for scaling (as we'd like the range [126.5,
  127.5) to map to the integer 127), but the actual clipping of float values
  needs to clip to just below 127.5, to something like 127.499999, such that
  RTNE rounds into range.

  Note: This is not necessary for SmallFloatConfig because the get_clip_bound
  returns the maximum possible value already.

  Args:
    config: the integer quantization config

  Returns:
    A value slightly below `get_clip_bound()`.
  """
  cb_unsafe = get_clip_bound(config)
  # It is essential that x is not clipped to cb_unsafe, but a smaller number.
  # Clipping to cb_unsafe moves the return value to a new and incorrect bucket.
  # The max supported (config.bits)-sized signed int is (2**(config.bits-1)-1).
  # The floor of cb_unsafe is larger than this limit, whereas using
  # cb_unsafe * (1 - eps) for some float32-representable eps >= 2**-23
  # accomplishes this. On the other hand, any eps above bucket resolution
  # 2**(-config.bits) would overcorrect, underutilizing integer range.
  cb = cb_unsafe - 2.0**(-20 + config.bits)
  assert cb < cb_unsafe, 'Internal error, epsilon too small.'
  return cb



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a reasonable baseline configuration for the WMT transformer.

Specific experimental configurations should override these values as needed.
"""

import enum

from aqt.jax_legacy.jax.wmt_mlperf.hparams_config_scripts import config_schema


class QuantTarget(enum.Enum):
  """Specifies which part of the model to quantize."""

  # No quantization
  NONE = enum.auto()

  # Quantize only the model weights. Activations are left unquantized.
  WEIGHTS_ONLY = enum.auto()

  # Quantize all the weights and activations. Activations use fixed
  # pre-specified clipping bounds.
  WEIGHTS_AND_FIXED_ACTS = enum.auto()

  # Quantize all the weights and activations. Activations use clipping bounds
  # derived from activation statistics collected during training.
  WEIGHTS_AND_AUTO_ACTS = enum.auto()


def get_base_config(n_layers: int, use_auto_acts: bool, fp_quant: bool):
  """Returns config that sets hyperparameters common to all quantization targets.

  Fields in that config can then be overridden to customize a configuration.

  Note that two hyperparameters, the number of layers and whether to
  automatically find clipping bounds for activations, have to be specified in
  advance as keyword arguments to this function instead of being overridden in
  the returned configdict. That is because these parameters affect the name and
  number of fields in the configdict instance, which can't be changed after
  creation: there will be one set of overridable parameters per layer in the
  configdict, and the field names in the 'quant_act' fields change depending on
  'use_auto_acts'.

  Args:
    n_layers: Number of layers in the encoder and decoder (eg, n_layers=3 mean
      three encoder layers and three decoder layers).
    use_auto_acts: Whether to use automatic bounds calculation for activations
      (True) or fixed bounds.
    fp_quant: Whether to use floating point quantization. Defaults to False for
      integer quantization.

  Returns:
    A ConfigDict instance suitable for WMT training.
  """
  config = config_schema.get_config(
      use_auto_acts=use_auto_acts, n_layers=n_layers, fp_quant=fp_quant)
  config.half_shift = False
  config.update({
      "learning_rate_schedule": {
          "factors": "constant * linear_warmup * rsqrt_decay",
          "base_learning_rate": 0.0625,
          "warmup_steps": 1000,
          "decay_factor": 0.5,
          "steps_per_decay": 20000,
          "steps_per_cycle": 100000,
      },
      "per_host_batch_size": 256,
      "num_train_steps": 200000,
      "weight_decay": 0.25,
      "beta1": 0.9,
      "beta2": 0.98,
      "eps": 1e-9,
      "random_seed": 0,
      "hardware_rng": True,
      "activation_bound_update_freq": -1,
      "activation_bound_start_step": -1,
      "weight_outlier_regularization": 0.0,
      "prefer_int8_to_int32_dot": True,
      "model_hparams": {
          "emb_dim": 1024,
          "num_heads": 16,
          "qkv_dim": 1024,
          "mlp_dim": 4096,
          "share_embeddings": True,
          "logits_via_embedding": True,
      },
      "weight_outlier_regularization_regex": "^.*kernel$",
      "weight_quant_granularity": "per_channel",

      "sparsity": {
          "type": "STRUCTURED_NM",
          "prune_rate": None,
          "order": "R",  # row wise sparsity
          "absolute": True,
          "smallest": True,
          "structure_decay": False,
          "sparse_ste": False
      },
  })
  config.act_sparsity.order = "C"
  if not fp_quant:
    config.prec = None
    config.quant_type = "aqt"
  else:
    config.prec.is_scaled = False
    config.quant_type = "fake_quant"
  layernorm_config = config_schema.get_layer_norm_config(
      quantized=False, quantized_reductions=False)
  config = config_schema.set_global_layer_norm_config(config, layernorm_config)
  return config


def get_auto_acts_config(n_layers: int, fp_quant: bool):
  """Returns config appropriate for automatic activation quantization."""
  config = get_base_config(
      n_layers=n_layers, use_auto_acts=True, fp_quant=fp_quant)
  config.quant_act.bounds.update({
      "initial_bound": -1.0,
      "stddev_coeff": 3.0,
      "absdev_coeff": 0.0,
      "mix_coeff": 1.0,
      "reset_stats": False,
      "ema_coeff": 0.1,
      "use_cams": False,
      "exclude_zeros": True,
      "use_mean_of_max": False,
      "granularity": "per_channel"
  })
  config.attention.quant_act.bounds.granularity = "per_tensor"
  config.weight_outlier_regularization = 1.0
  config.activation_bound_start_step = 20000
  return config


def get_weights_only_config(n_layers: int, fp_quant: bool):
  """Returns config for weights-only quantization."""
  config = get_base_config(
      n_layers=n_layers, use_auto_acts=False, fp_quant=fp_quant)
  config.quant_act.bounds = None
  return config


def get_fixed_acts_config(n_layers: int, fp_quant: bool):
  """Returns config for activation quantization with fixed bounds."""
  config = get_base_config(
      n_layers=n_layers, use_auto_acts=False, fp_quant=fp_quant)
  config.quant_act.bounds = 1.0
  config.mlp_block.dense_2.quant_act.bounds = 6.0
  config.attention.attn_acts.attn_act_v.bounds = 6.0
  config.attention.dense_out.quant_act.bounds = 2.0
  return config


def get_config(quant_target: QuantTarget,
               n_layers: int,
               fp_quant: bool = False):
  """Returns config for a given quantization target and layer count."""
  if quant_target == QuantTarget.WEIGHTS_AND_AUTO_ACTS:
    return get_auto_acts_config(n_layers=n_layers, fp_quant=fp_quant)
  elif quant_target == QuantTarget.WEIGHTS_ONLY:
    return get_weights_only_config(n_layers=n_layers, fp_quant=fp_quant)
  elif quant_target == QuantTarget.WEIGHTS_AND_FIXED_ACTS:
    return get_fixed_acts_config(n_layers=n_layers, fp_quant=fp_quant)
  elif quant_target == QuantTarget.NONE:
    return get_base_config(
        n_layers=n_layers, use_auto_acts=False, fp_quant=fp_quant)
  else:
    raise ValueError(f"quant_target {quant_target} not understood.")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Contains the hparams class used for ResNet."""

import dataclasses
import typing
from typing import Any

from aqt.jax_legacy.jax.flax import struct as flax_struct
from aqt.jax_legacy.utils import hparams_utils

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass


@dataclass
class LrScheduler:
  warmup_epochs: int
  cooldown_epochs: int
  scheduler: str
  num_epochs: int
  endlr: float
  knee_lr: float
  knee_epochs: int


@dataclass
class Adam:
  beta1: float
  beta2: float


@dataclass
class TrainingHParams:
  """Hyperparameters used for training."""

  # Metadata
  metadata: hparams_utils.HParamsMetadata

  # General hparams
  base_learning_rate: float
  momentum: float  # only used when optimier=='sgd'
  weight_decay: float
  lr_scheduler: LrScheduler
  optimizer: str
  adam: Adam  # only used when optimizer=='adam'
  early_stop_steps: int
  teacher_model: str
  is_teacher: bool
  seed: int

  # Auto-clip activation quantization hparams. See
  # train_utils.should_update_bounds for more details. We use -1 instead of None
  # to indicate that these parameters should be inactive since if these
  # attributes are marked as optional, Dacite will set them to None even if
  # they are missing from the JSON-serialized version of this dataclass, while
  # we want Dacite to raise an error in this case to alert users about an
  # incomplete configuration.
  activation_bound_update_freq: int
  activation_bound_start_step: int
  weight_quant_start_step: int

  # Model hparams
  model_hparams: Any



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT pax."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2.pax import pax_base_ops
import jax
import jax.numpy as jnp


class PaxBaseOpsTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(cfg=None),
      dict(cfg=config.config_v4()),
  )
  def test_einsum_is_quantized(self, cfg):
    prng_key = jax.random.PRNGKey(seed=123)
    rngs = {'params': prng_key, 'random': prng_key}
    lhs = jax.random.normal(prng_key, [10, 10, 10])
    rhs = jax.random.normal(prng_key, [10, 10])
    eqn = '...x,xy->y'

    layer = pax_base_ops.AqtEinsum(cfg=cfg, track_train_step=True)
    variable = layer.init(rngs, eqn, lhs, rhs)

    @jax.jit
    def apply_fn(lhs, rhs):
      return layer.apply(variable, eqn, lhs, rhs, mutable=True, rngs=rngs)

    @jax.jit
    @jax.value_and_grad
    def train_step(lhs, rhs):
      return jnp.sum(apply_fn(lhs, rhs)[0])

    # Run one value and grad to test if training will work
    train_step(lhs, rhs)

    out_float = jnp.einsum(eqn, lhs, rhs)
    out_quant, new_variable = apply_fn(lhs, rhs)
    if cfg is None:
      # Ensure results are the same
      self.assertEqual(jnp.mean(jnp.square(out_float - out_quant)), 0)
    else:
      # Ensure quantized einsum does not produce the same result.
      self.assertGreater(jnp.mean(jnp.square(out_float - out_quant)), 0)

    self.assertEqual(new_variable['non_trainable']['train_step'], 1)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test for AQT flax."""
import functools
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import config
from aqt.jax.v2.flax import aqt_flax
import flax.linen as nn
import jax
import jax.numpy as jnp


class AqtFlaxTest(parameterized.TestCase):

  def test_aqt_einsum(self):
    class Model(nn.Module):
      aqt_cfg: config.DotGeneral | None
      lhs_qt_external: bool = False
      rhs_qt_external: bool = False

      @nn.compact
      def __call__(self, lhs, rhs):
        lhs_in = lhs
        assert not (
            (self.lhs_qt_external or self.rhs_qt_external)
            and (self.aqt_cfg is None)
        ), 'aqt_cfg cannot be None when providing qtensor as inputs to einsum'
        if self.lhs_qt_external:
          lhs_in, _ = self.aqt_cfg.fwd.lhs.quantizer.quant(
              lhs, calibration_axes=(2, 3)
          )
          lhs_dtype = self.aqt_cfg.fwd.lhs.quantizer.numerics.get_dtype()
          lhs_in = lhs_in.qvalue_astype(lhs_dtype)
        rhs_in = rhs
        if self.rhs_qt_external:
          rhs_in, _ = self.aqt_cfg.fwd.rhs.quantizer.quant(
              rhs, calibration_axes=(1, 2)
          )
          rhs_dtype = self.aqt_cfg.fwd.rhs.quantizer.numerics.get_dtype()
          rhs_in = rhs_in.qvalue_astype(rhs_dtype)
        einsum = aqt_flax.AqtEinsum(cfg=self.aqt_cfg)
        # xhs_qt can be inputs to AqtEinsum
        # xhs->xhs_qt can happen outside of AqtEinsum, e.g., k/v cache quant
        # input xhs_qt will force get_tensor() to always return xhs_qt
        out = einsum('ijkh,mkh->ijm', lhs_in, rhs_in)
        return out

    key = jax.random.PRNGKey(0)
    subkey1, subkey2 = jax.random.split(key, num=2)
    lhs = jax.random.uniform(subkey1, shape=(3, 4, 5, 6))
    rhs = jax.random.uniform(subkey2, shape=(2, 5, 6))

    def test(model_cls, cfg):
      model = model_cls(cfg)
      out = model.apply({}, lhs, rhs, rngs={'params': jax.random.PRNGKey(0)})
      # print(f'{out.shape=}')
      # print(f'{out=}')
      return out

    out_float = test(Model, None)
    out_int8 = test(Model, config.config_v4())
    out_int8_lqt = test(
        functools.partial(Model, lhs_qt_external=True), config.config_v4()
    )
    out_int8_rqt = test(
        functools.partial(Model, rhs_qt_external=True), config.config_v4()
    )
    out_int8_qt = test(
        functools.partial(Model, lhs_qt_external=True, rhs_qt_external=True),
        config.config_v4(),
    )

    assert (out_int8 == out_int8_lqt).all(), 'lhs external qt failed'
    assert (out_int8 == out_int8_rqt).all(), 'rhs external qt failed'
    assert (out_int8 == out_int8_qt).all(), 'both external qt failed'
    mse = jnp.mean(jnp.square(out_int8_qt - out_float))
    assert mse > 0, 'Mean square error is 0. Einsum is not quantized.'

  def test_einsum_grad_leak(self):
    class CNN(nn.Module):
      aqt_cfg: config.DotGeneral

      @nn.compact
      def __call__(self, x):
        einsum = aqt_flax.AqtEinsum(self.aqt_cfg)
        x = einsum('bc,ab->ac', jnp.identity(10, dtype=x.dtype), x)
        return x

    model = CNN(aqt_cfg=config.fully_quantized())
    var = model.init({'params': jax.random.PRNGKey(0)}, jnp.ones(shape=(1, 10)))

    @jax.jit
    def apply_fn(inputs):
      return model.apply(
          var, inputs, rngs={'params': jax.random.PRNGKey(0)}, mutable=True
      )

    @jax.jit
    @jax.value_and_grad
    def train_step(inputs):
      return jnp.sum(apply_fn(inputs)[0])

    train_step(jnp.ones(shape=(1, 10)))


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Shape utility functions used by the models."""

from typing import Sequence


def assert_shapes_equal(shape: Sequence[int],
                        expected_shape: Sequence[int]) -> None:
  """Check if the shape of a tensor exactly matches an expected shape."""

  assert shape == expected_shape, (f"Expected tensor to have shape "
                                   f"{expected_shape} but got {shape}.")


def assert_shapes_compatible(lhs: Sequence[int], rhs: Sequence[int]):
  """Check if the shape of two tensors are broadcast-compatible."""

  # This is a stricter-than-necessary check for broadcast-comptability,
  # but it's error-prone to allow broadcasting to insert new dimensions so
  # we don't allow that.
  if len(lhs) != len(rhs):
    return False

  for lhs_dim, rhs_dim in zip(lhs, rhs):
    # A dimension of 1 will be broadcasted to match the other tensor's dimension
    if lhs_dim == 1 or rhs_dim == 1:
      continue
    assert lhs_dim == rhs_dim, (f"Tensors with shape {lhs} and {rhs} cannot "
                                f"be broadcasted together")



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for imagenet.configs."""

from absl.testing import absltest
from absl.testing import parameterized

from aqt.jax_legacy.jax.imagenet.configs import resnet101_bfloat16
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w4_a2_fixed
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w8
from aqt.jax_legacy.jax.imagenet.configs import resnet50_w8_a8_fixed
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_bfloat16
from aqt.jax_legacy.jax.imagenet.configs.paper import resnet50_w4_a4_auto


class ConfigsTest(parameterized.TestCase):

  @parameterized.parameters(
      dict(config_dict_module=resnet50_bfloat16),
      dict(config_dict_module=resnet101_bfloat16))
  def test_unquantized_config(self, config_dict_module):
    config = config_dict_module.get_config()
    self.assertIsNone(config.prec)

  @parameterized.parameters(
      dict(config_dict_module=resnet50_w8, prec=8),
      dict(config_dict_module=resnet50_w4, prec=4))
  def test_weights_only_quantized_config(self, config_dict_module, prec):
    config = config_dict_module.get_config()
    self.assertEqual(config.weight_prec, prec)
    self.assertIsNone(config.quant_act.bounds)

  @parameterized.parameters(
      dict(config_dict_module=resnet50_w4_a4_auto, weight_prec=4, act_prec=4),
      dict(config_dict_module=resnet50_w4_a2_fixed, weight_prec=4, act_prec=2),
      dict(config_dict_module=resnet50_w8_a8_fixed, weight_prec=8, act_prec=8))
  def test_fully_quantized_quantized_config(self, config_dict_module,
                                            weight_prec, act_prec):
    config = config_dict_module.get_config()
    self.assertEqual(config.weight_prec, weight_prec)
    self.assertEqual(config.quant_act.prec, act_prec)


if __name__ == '__main__':
  absltest.main()
